"""
LUT Processing
==============

Defines the classes and definitions handling *LUT* processing:

-   :class:`colour.LUT1D`
-   :class:`colour.LUT3x1D`
-   :class:`colour.LUT3D`
-   :class:`colour.io.LUT_to_LUT`
"""

from __future__ import annotations

import numpy as np
from abc import ABC, abstractmethod
from copy import deepcopy
from operator import (
    add,
    mul,
    pow,
    sub,
    truediv,
    iadd,
    imul,
    ipow,
    isub,
    itruediv,
)

from colour.algebra import (
    Extrapolator,
    LinearInterpolator,
    linear_conversion,
    table_interpolation_trilinear,
)
from colour.hints import (
    Any,
    ArrayLike,
    Boolean,
    FloatingOrArrayLike,
    Integer,
    IntegerOrArrayLike,
    List,
    Literal,
    NDArray,
    Optional,
    Sequence,
    Type,
    Union,
    cast,
)
from colour.utilities import (
    as_float_array,
    as_int,
    as_int_array,
    as_int_scalar,
    attest,
    is_numeric,
    is_iterable,
    is_string,
    full,
    optional,
    required,
    runtime_warning,
    tsplit,
    tstack,
    usage_warning,
    validate_method,
)
from colour.utilities.deprecation import ObjectRenamed

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "AbstractLUT",
    "LUT1D",
    "LUT3x1D",
    "LUT3D",
    "LUT_to_LUT",
]


class AbstractLUT(ABC):
    """
    Define the base class for *LUT*.

    This is an :class:`ABCMeta` abstract class that must be inherited by
    sub-classes.

    Parameters
    ----------
    table
        Underlying *LUT* table.
    name
        *LUT* name.
    dimensions
        *LUT* dimensions, typically, 1 for a 1D *LUT*, 2 for a 3x1D *LUT* and 3
        for a 3D *LUT*.
    domain
        *LUT* domain, also used to define the instantiation time default table
        domain.
    size
        *LUT* size, also used to define the instantiation time default table
        size.
    comments
        Comments to add to the *LUT*.

    Attributes
    ----------
    -   :attr:`~colour.io.luts.lut.AbstractLUT.table`
    -   :attr:`~colour.io.luts.lut.AbstractLUT.name`
    -   :attr:`~colour.io.luts.lut.AbstractLUT.dimensions`
    -   :attr:`~colour.io.luts.lut.AbstractLUT.domain`
    -   :attr:`~colour.io.luts.lut.AbstractLUT.size`
    -   :attr:`~colour.io.luts.lut.AbstractLUT.comments`

    Methods
    -------
    -   :meth:`~colour.io.luts.lut.AbstractLUT.__init__`
    -   :meth:`~colour.io.luts.lut.AbstractLUT.__str__`
    -   :meth:`~colour.io.luts.lut.AbstractLUT.__repr__`
    -   :meth:`~colour.io.luts.lut.AbstractLUT.__eq__`
    -   :meth:`~colour.io.luts.lut.AbstractLUT.__ne__`
    -   :meth:`~colour.io.luts.lut.AbstractLUT.__add__`
    -   :meth:`~colour.io.luts.lut.AbstractLUT.__iadd__`
    -   :meth:`~colour.io.luts.lut.AbstractLUT.__sub__`
    -   :meth:`~colour.io.luts.lut.AbstractLUT.__isub__`
    -   :meth:`~colour.io.luts.lut.AbstractLUT.__mul__`
    -   :meth:`~colour.io.luts.lut.AbstractLUT.__imul__`
    -   :meth:`~colour.io.luts.lut.AbstractLUT.__div__`
    -   :meth:`~colour.io.luts.lut.AbstractLUT.__idiv__`
    -   :meth:`~colour.io.luts.lut.AbstractLUT.__pow__`
    -   :meth:`~colour.io.luts.lut.AbstractLUT.__ipow__`
    -   :meth:`~colour.io.luts.lut.AbstractLUT.arithmetical_operation`
    -   :meth:`~colour.io.luts.lut.AbstractLUT.is_domain_explicit`
    -   :meth:`~colour.io.luts.lut.AbstractLUT.linear_table`
    -   :meth:`~colour.io.luts.lut.AbstractLUT.copy`
    -   :meth:`~colour.io.luts.lut.AbstractLUT.invert`
    -   :meth:`~colour.io.luts.lut.AbstractLUT.apply`
    -   :meth:`~colour.io.luts.lut.AbstractLUT.convert`
    """

    def __init__(
        self,
        table: Optional[ArrayLike] = None,
        name: Optional[str] = None,
        dimensions: Optional[Integer] = None,
        domain: Optional[ArrayLike] = None,
        size: Optional[IntegerOrArrayLike] = None,
        comments: Optional[Sequence] = None,
    ):
        self._name: str = f"Unity {size!r}" if table is None else f"{id(self)}"
        self.name = optional(name, self._name)
        self._dimensions = optional(dimensions, 0)
        self._table: NDArray = self.linear_table(
            cast(ArrayLike, optional(size, 0)),
            cast(ArrayLike, optional(domain, np.array([]))),
        )
        self.table = cast(
            ArrayLike, optional(table, self._table)
        )  # type: ignore[assignment]
        # TODO: Remove pragma when https://github.com/python/mypy/issues/3004
        # is resolved.
        self._domain: NDArray = np.array([])
        self.domain = cast(
            ArrayLike, optional(domain, self._domain)
        )  # type: ignore[assignment]
        self._comments: List = []
        self.comments = cast(
            ArrayLike, optional(comments, self._comments)
        )  # type: ignore[assignment]

    @property
    def table(self) -> NDArray:
        """
        Getter and setter property for the underlying *LUT* table.

        Parameters
        ----------
        value
            Value to set the underlying *LUT* table with.

        Returns
        -------
        :class:`numpy.ndarray`
            Underlying *LUT* table.
        """

        return self._table

    @table.setter
    def table(self, value: ArrayLike):
        """Setter for the **self.table** property."""

        self._table = self._validate_table(value)

    @property
    def name(self) -> str:
        """
        Getter and setter property for the *LUT* name.

        Parameters
        ----------
        value
            Value to set the *LUT* name with.

        Returns
        -------
        :class:`str`
            *LUT* name.
        """

        return self._name

    @name.setter
    def name(self, value: str):
        """Setter for the **self.name** property."""

        attest(
            is_string(value),
            f'"name" property: "{value}" type is not "str"!',
        )

        self._name = value

    @property
    def domain(self) -> NDArray:
        """
        Getter and setter property for the *LUT* domain.

        Parameters
        ----------
        value
            Value to set the *LUT* domain with.

        Returns
        -------
        :class:`numpy.ndarray`
            *LUT* domain.
        """

        return self._domain

    @domain.setter
    def domain(self, value: ArrayLike):
        """Setter for the **self.domain** property."""

        # pylint: disable=E1121
        self._domain = self._validate_domain(value)

    @property
    def dimensions(self) -> Integer:
        """
        Getter property for the *LUT* dimensions.

        Returns
        -------
        :class:`numpy.integer`
            *LUT* dimensions.
        """

        return self._dimensions

    @property
    def size(self) -> Integer:
        """
        Getter property for the *LUT* size.

        Returns
        -------
        :class:`numpy.integer`
            *LUT* size.
        """

        return self._table.shape[0]

    @property
    def comments(self) -> List:
        """
        Getter and setter property for the *LUT* comments.

        Parameters
        ----------
        value
            Value to set the *LUT* comments with.

        Returns
        -------
        :class:`list`
            *LUT* comments.
        """

        return self._comments

    @comments.setter
    def comments(self, value: Sequence):
        """Setter for the **self.comments** property."""

        attest(
            is_iterable(value),
            f'"comments" property: "{value}" must be a sequence!',
        )

        self._comments = list(value)

    def __str__(self) -> str:
        """
        Return a formatted string representation of the *LUT*.

        Returns
        -------
        :class:`str`
            Formatted string representation.
        """

        def _indent_array(a: ArrayLike) -> str:
            """Indent given array string representation."""

            return str(a).replace(" [", " " * 14 + "[")

        comments = "\n".join(
            [
                f"Comment {str(i + 1).zfill(2)} : {comment}"
                for i, comment in enumerate(self.comments)
            ]
        )
        comments = f"\n{comments}" if comments else ""

        underline = "-" * (len(self.__class__.__name__) + 3 + len(self.name))

        return (
            f"{self.__class__.__name__} - {self.name}\n"
            f"{underline}\n\n"
            f"Dimensions : {self.dimensions}\n"
            f"Domain     : {_indent_array(self.domain)}\n"
            f'Size       : {str(self.table.shape).replace("L", "")}'
            f"{comments}"
        )

    def __repr__(self) -> str:
        """
        Return an evaluable string representation of the *LUT*.

        Returns
        -------
        :class:`str`
            Evaluable string representation.
        """

        representation = repr(self.table)
        representation = representation.replace(
            "array", self.__class__.__name__
        )
        representation = representation.replace(
            "       [", f"{' ' * (len(self.__class__.__name__) + 2)}["
        )

        domain = repr(self.domain).replace("array(", "").replace(")", "")
        domain = domain.replace(
            "       [", f"{' ' * (len(self.__class__.__name__) + 9)}["
        )

        indentation = " " * (len(self.__class__.__name__) + 1)
        comments = (
            f",\n\n{indentation}comments={repr(self.comments)}"
            if self.comments
            else ""
        )

        return (
            f"{representation[:-1]},\n"
            f"{indentation}name='{self.name}',\n"
            f"{indentation}domain={domain}"
            f"{comments})"
        )

    def __eq__(self, other: Any) -> bool:
        """
        Return whether the *LUT* is equal to given other object.

        Parameters
        ----------
        other
            Object to test whether it is equal to the *LUT*.

        Returns
        -------
        :class:`bool`
            Whether given object is equal to the *LUT*.
        """

        if isinstance(other, AbstractLUT):
            if all(
                [
                    np.array_equal(self.table, other.table),
                    np.array_equal(self.domain, other.domain),
                ]
            ):
                return True

        return False

    def __ne__(self, other: Any) -> bool:
        """
        Return whether the *LUT* is not equal to given other object.

        Parameters
        ----------
        other
            Object to test whether it is not equal to the *LUT*.

        Returns
        -------
        :class:`bool`
            Whether given object is not equal to the *LUT*.
        """

        return not (self == other)

    def __add__(
        self, a: Union[FloatingOrArrayLike, AbstractLUT]
    ) -> AbstractLUT:
        """
        Implement support for addition.

        Parameters
        ----------
        a
            :math:`a` variable to add.

        Returns
        -------
        :class:`colour.io.luts.lut.AbstractLUT`
            Variable added *LUT*.
        """

        return self.arithmetical_operation(a, "+")

    def __iadd__(
        self, a: Union[FloatingOrArrayLike, AbstractLUT]
    ) -> AbstractLUT:
        """
        Implement support for in-place addition.

        Parameters
        ----------
        a
            :math:`a` variable to add in-place.

        Returns
        -------
        :class:`colour.io.luts.lut.AbstractLUT`
            In-place variable added *LUT*.
        """

        return self.arithmetical_operation(a, "+", True)

    def __sub__(
        self, a: Union[FloatingOrArrayLike, AbstractLUT]
    ) -> AbstractLUT:
        """
        Implement support for subtraction.

        Parameters
        ----------
        a
            :math:`a` variable to subtract.

        Returns
        -------
        :class:`colour.io.luts.lut.AbstractLUT`
            Variable subtracted *LUT*.
        """

        return self.arithmetical_operation(a, "-")

    def __isub__(
        self, a: Union[FloatingOrArrayLike, AbstractLUT]
    ) -> AbstractLUT:
        """
        Implement support for in-place subtraction.

        Parameters
        ----------
        a
            :math:`a` variable to subtract in-place.

        Returns
        -------
        :class:`colour.io.luts.lut.AbstractLUT`
            In-place variable subtracted *LUT*.
        """

        return self.arithmetical_operation(a, "-", True)

    def __mul__(
        self, a: Union[FloatingOrArrayLike, AbstractLUT]
    ) -> AbstractLUT:
        """
        Implement support for multiplication.

        Parameters
        ----------
        a
            :math:`a` variable to multiply by.

        Returns
        -------
        :class:`colour.io.luts.lut.AbstractLUT`
            Variable multiplied *LUT*.
        """

        return self.arithmetical_operation(a, "*")

    def __imul__(
        self, a: Union[FloatingOrArrayLike, AbstractLUT]
    ) -> AbstractLUT:
        """
        Implement support for in-place multiplication.

        Parameters
        ----------
        a
            :math:`a` variable to multiply by in-place.

        Returns
        -------
        :class:`colour.io.luts.lut.AbstractLUT`
            In-place variable multiplied *LUT*.
        """

        return self.arithmetical_operation(a, "*", True)

    def __div__(
        self, a: Union[FloatingOrArrayLike, AbstractLUT]
    ) -> AbstractLUT:
        """
        Implement support for division.

        Parameters
        ----------
        a
            :math:`a` variable to divide by.

        Returns
        -------
        :class:`colour.io.luts.lut.AbstractLUT`
            Variable divided *LUT*.
        """

        return self.arithmetical_operation(a, "/")

    def __idiv__(
        self, a: Union[FloatingOrArrayLike, AbstractLUT]
    ) -> AbstractLUT:
        """
        Implement support for in-place division.

        Parameters
        ----------
        a
            :math:`a` variable to divide by in-place.

        Returns
        -------
        :class:`colour.io.luts.lut.AbstractLUT`
            In-place variable divided *LUT*.
        """

        return self.arithmetical_operation(a, "/", True)

    __itruediv__ = __idiv__
    __truediv__ = __div__

    def __pow__(
        self, a: Union[FloatingOrArrayLike, AbstractLUT]
    ) -> AbstractLUT:
        """
        Implement support for exponentiation.

        Parameters
        ----------
        a
            :math:`a` variable to exponentiate by.

        Returns
        -------
        :class:`colour.io.luts.lut.AbstractLUT`
            Variable exponentiated *LUT*.
        """

        return self.arithmetical_operation(a, "**")

    def __ipow__(
        self, a: Union[FloatingOrArrayLike, AbstractLUT]
    ) -> AbstractLUT:
        """
        Implement support for in-place exponentiation.

        Parameters
        ----------
        a
            :math:`a` variable to exponentiate by in-place.

        Returns
        -------
        :class:`colour.io.luts.lut.AbstractLUT`
            In-place variable exponentiated *LUT*.
        """

        return self.arithmetical_operation(a, "**", True)

    def arithmetical_operation(
        self,
        a: Union[FloatingOrArrayLike, AbstractLUT],
        operation: Literal["+", "-", "*", "/", "**"],
        in_place: Boolean = False,
    ) -> AbstractLUT:
        """
        Perform given arithmetical operation with :math:`a` operand, the
        operation can be either performed on a copy or in-place, must be
        reimplemented by sub-classes.

        Parameters
        ----------
        a
            Operand.
        operation
            Operation to perform.
        in_place
            Operation happens in place.

        Returns
        -------
        :class:`colour.io.luts.lut.AbstractLUT`
            *LUT*.
        """

        operator, ioperator = {
            "+": (add, iadd),
            "-": (sub, isub),
            "*": (mul, imul),
            "/": (truediv, itruediv),
            "**": (pow, ipow),
        }[operation]

        if in_place:
            if isinstance(a, AbstractLUT):
                operand = a.table
            else:
                operand = as_float_array(a)

            self.table = operator(self.table, operand)

            return self
        else:
            copy = ioperator(self.copy(), a)

            return copy

    @abstractmethod
    def _validate_table(self, table: ArrayLike) -> NDArray:
        """
        Validate given table according to *LUT* dimensions.

        Parameters
        ----------
        table
            Table to validate.

        Returns
        -------
        :class:`numpy.ndarray`
            Validated table as a :class:`ndarray` instance.
        """

        pass

    @abstractmethod
    def _validate_domain(self, domain: ArrayLike) -> NDArray:
        """
        Validate given domain according to *LUT* dimensions.

        Parameters
        ----------
        domain
            Domain to validate.

        Returns
        -------
        :class:`numpy.ndarray`
            Validated domain as a :class:`ndarray` instance.
        """

        pass

    @abstractmethod
    def is_domain_explicit(self) -> Boolean:
        """
        Return whether the *LUT* domain is explicit (or implicit).

        An implicit domain is defined by its shape only::

            [[0 1]
             [0 1]
             [0 1]]

        While an explicit domain defines every single discrete samples::

            [[0.0 0.0 0.0]
             [0.1 0.1 0.1]
             [0.2 0.2 0.2]
             [0.3 0.3 0.3]
             [0.4 0.4 0.4]
             [0.8 0.8 0.8]
             [1.0 1.0 1.0]]

        Returns
        -------
        :class:`bool`
            Is *LUT* domain explicit.
        """

        pass

    @staticmethod
    @abstractmethod
    def linear_table(
        size: Optional[IntegerOrArrayLike] = None,
        domain: Optional[ArrayLike] = None,
    ) -> NDArray:
        """
        Return a linear table of given size according to *LUT* dimensions.

        Parameters
        ----------
        size
            Expected table size, for a 1D *LUT*, the number of output samples
            :math:`n` is equal to ``size``, for a 3x1D *LUT* :math:`n` is equal
            to ``size * 3`` or ``size[0] + size[1] + size[2]``, for a 3D *LUT*
            :math:`n` is equal to ``size**3 * 3`` or
            ``size[0] * size[1] * size[2] * 3``.
        domain
            Domain of the table.

        Returns
        -------
        :class:`numpy.ndarray`
            Linear table.
        """

        pass

    def copy(self) -> AbstractLUT:
        """
        Return a copy of the sub-class instance.

        Returns
        -------
        :class:`colour.io.luts.lut.AbstractLUT`
            *LUT* copy.
        """

        return deepcopy(self)

    @abstractmethod
    def invert(self, **kwargs: Any) -> AbstractLUT:
        """
        Compute and returns an inverse copy of the *LUT*.

        Other Parameters
        ----------------
        kwargs
            Keywords arguments.

        Returns
        -------
        :class:`colour.io.luts.lut.AbstractLUT`
            Inverse *LUT* class instance.
        """

        pass

    @abstractmethod
    def apply(self, RGB: ArrayLike, **kwargs: Any) -> NDArray:
        """
        Apply the *LUT* to given *RGB* colourspace array using given method.

        Parameters
        ----------
        RGB
            *RGB* colourspace array to apply the *LUT* onto.

        Other Parameters
        ----------------
        direction
            Whether the *LUT* should be applied in the forward or inverse
            direction.
        extrapolator
            Extrapolator class type or object to use as extrapolating function.
        extrapolator_kwargs
            Arguments to use when instantiating or calling the extrapolating
            function.
        interpolator
            Interpolator class type or object to use as interpolating function.
        interpolator_kwargs
            Arguments to use when instantiating or calling the interpolating
            function.

        Returns
        -------
        :class:`numpy.ndarray`
            Interpolated *RGB* colourspace array.
        """

        pass

    @abstractmethod
    def convert(
        self,
        cls: Type[AbstractLUT],
        force_conversion: Boolean = False,
        **kwargs: Any,
    ) -> AbstractLUT:
        """
        Convert the *LUT* to given ``cls`` class instance.

        Parameters
        ----------
        cls
            *LUT* class instance.
        force_conversion
            Whether to force the conversion as it might be destructive.

        Other Parameters
        ----------------
        interpolator
            Interpolator class type to use as interpolating function.
        interpolator_kwargs
            Arguments to use when instantiating the interpolating function.
        size
            Expected table size in case of an upcast to or a downcast from a
            :class:`LUT3D` class instance.

        Returns
        -------
        :class:`colour.io.luts.lut.AbstractLUT`
            Converted *LUT* class instance.

        Warnings
        --------
        Some conversions are destructive and raise a :class:`ValueError`
        exception by default.

        Raises
        ------
        ValueError
            If the conversion is destructive.
        """

        pass


class LUT1D(AbstractLUT):
    """
    Define the base class for a 1D *LUT*.

    Parameters
    ----------
    table
        Underlying *LUT* table.
    name
        *LUT* name.
    domain
        *LUT* domain, also used to define the instantiation time default table
        domain.
    size
        Size of the instantiation time default table, default to 10.
    comments
        Comments to add to the *LUT*.

    Methods
    -------
    -   :meth:`~colour.LUT1D.__init__`
    -   :meth:`~colour.LUT1D.is_domain_explicit`
    -   :meth:`~colour.LUT1D.linear_table`
    -   :meth:`~colour.LUT1D.invert`
    -   :meth:`~colour.LUT1D.apply`
    -   :meth:`~colour.LUT1D.convert`

    Examples
    --------
    Instantiating a unity LUT with a table with 16 elements:

    >>> print(LUT1D(size=16))
    LUT1D - Unity 16
    ----------------
    <BLANKLINE>
    Dimensions : 1
    Domain     : [ 0.  1.]
    Size       : (16,)

    Instantiating a LUT using a custom table with 16 elements:

    >>> print(LUT1D(LUT1D.linear_table(16) ** (1 / 2.2)))  # doctest: +ELLIPSIS
    LUT1D - ...
    --------...
    <BLANKLINE>
    Dimensions : 1
    Domain     : [ 0.  1.]
    Size       : (16,)

    Instantiating a LUT using a custom table with 16 elements, custom name,
    custom domain and comments:

    >>> from colour.algebra import spow
    >>> domain = np.array([-0.1, 1.5])
    >>> print(LUT1D(
    ...     spow(LUT1D.linear_table(16, domain), 1 / 2.2),
    ...     'My LUT',
    ...     domain,
    ...     comments=['A first comment.', 'A second comment.']))
    LUT1D - My LUT
    --------------
    <BLANKLINE>
    Dimensions : 1
    Domain     : [-0.1  1.5]
    Size       : (16,)
    Comment 01 : A first comment.
    Comment 02 : A second comment.
    """

    def __init__(
        self,
        table: Optional[ArrayLike] = None,
        name: Optional[str] = None,
        domain: Optional[ArrayLike] = None,
        size: Optional[IntegerOrArrayLike] = None,
        comments: Optional[Sequence] = None,
    ):

        domain = as_float_array(
            cast(ArrayLike, optional(domain, np.array([0, 1])))
        )
        size = cast(Integer, optional(size, 10))

        super().__init__(table, name, 1, domain, size, comments)

    def _validate_table(self, table: ArrayLike) -> NDArray:
        """
        Validate given table is a 1D array.

        Parameters
        ----------
        table
            Table to validate.

        Returns
        -------
        :class:`numpy.ndarray`
            Validated table as a :class:`ndarray` instance.
        """

        table = as_float_array(table)

        attest(len(table.shape) == 1, "The table must be a 1D array!")

        return table

    def _validate_domain(self, domain: ArrayLike) -> NDArray:
        """
        Validate given domain.

        Parameters
        ----------
        domain
            Domain to validate.

        Returns
        -------
        :class:`numpy.ndarray`
            Validated domain as a :class:`ndarray` instance.
        """

        domain = as_float_array(domain)

        attest(len(domain.shape) == 1, "The domain must be a 1D array!")

        attest(
            domain.shape[0] >= 2,
            "The domain column count must be equal or greater than 2!",
        )

        return domain

    def is_domain_explicit(self) -> Boolean:
        """
        Return whether the *LUT* domain is explicit (or implicit).

        An implicit domain is defined by its shape only::

            [0 1]

        While an explicit domain defines every single discrete samples::

            [0.0 0.1 0.2 0.4 0.8 1.0]

        Returns
        -------
        :class:`bool`
            Is *LUT* domain explicit.

        Examples
        --------
        >>> LUT1D().is_domain_explicit()
        False
        >>> table = domain = np.linspace(0, 1, 10)
        >>> LUT1D(table, domain=domain).is_domain_explicit()
        True
        """

        return len(self.domain) != 2

    @staticmethod
    def linear_table(
        size: Optional[IntegerOrArrayLike] = None,
        domain: Optional[ArrayLike] = None,
    ) -> NDArray:
        """
        Return a linear table, the number of output samples :math:`n` is equal
        to ``size``.

        Parameters
        ----------
        size
            Expected table size, default to 10.
        domain
            Domain of the table.

        Returns
        -------
        :class:`numpy.ndarray`
            Linear table with ``size`` samples.

        Examples
        --------
        >>> LUT1D.linear_table(5, np.array([-0.1, 1.5]))
        array([-0.1,  0.3,  0.7,  1.1,  1.5])
        >>> LUT1D.linear_table(domain=np.linspace(-0.1, 1.5, 5))
        array([-0.1,  0.3,  0.7,  1.1,  1.5])
        """

        size = cast(Integer, optional(size, 10))
        domain = as_float_array(
            cast(ArrayLike, optional(domain, np.array([0, 1])))
        )

        if len(domain) != 2:
            return domain
        else:
            attest(is_numeric(size), "Linear table size must be a numeric!")

            return np.linspace(domain[0], domain[1], as_int_scalar(size))

    def invert(self, **kwargs: Any) -> LUT1D:
        """
        Compute and returns an inverse copy of the *LUT*.

        Other Parameters
        ----------------
        kwargs
            Keywords arguments, only given for signature compatibility with
            the :meth:`AbstractLUT.invert` method.

        Returns
        -------
        :class:`colour.LUT1D`
            Inverse *LUT* class instance.

        Examples
        --------
        >>> LUT = LUT1D(LUT1D.linear_table() ** (1 / 2.2))
        >>> print(LUT.table)  # doctest: +ELLIPSIS
        [ 0.       ...  0.3683438...  0.5047603...  0.6069133...  \
0.6916988...  0.7655385...
          0.8316843...  0.8920493...  0.9478701...  1.        ]
        >>> print(LUT.invert())  # doctest: +ELLIPSIS
        LUT1D - ... - Inverse
        --------...----------
        <BLANKLINE>
        Dimensions : 1
        Domain     : [ 0.          0.3683438...  0.5047603...  0.6069133...  \
0.6916988...  0.7655385...
          0.8316843...  0.8920493...  0.9478701...  1.        ]
        Size       : (10,)
        >>> print(LUT.invert().table)  # doctest: +ELLIPSIS
        [ 0.       ...  0.1111111...  0.2222222...  0.3333333...  \
0.4444444...  0.5555555...
          0.6666666...  0.7777777...  0.8888888...  1.        ]
        """

        if self.is_domain_explicit():
            domain = self.domain
        else:
            domain_min, domain_max = self.domain
            domain = np.linspace(domain_min, domain_max, self.size)

        LUT_i = LUT1D(
            table=domain,
            name=f"{self.name} - Inverse",
            domain=self.table,
        )

        return LUT_i

    def apply(self, RGB: ArrayLike, **kwargs: Any) -> NDArray:
        """
        Apply the *LUT* to given *RGB* colourspace array using given method.

        Parameters
        ----------
        RGB
            *RGB* colourspace array to apply the *LUT* onto.

        Other Parameters
        ----------------
        direction
            Whether the *LUT* should be applied in the forward or inverse
            direction.
        extrapolator
            Extrapolator class type or object to use as extrapolating function.
        extrapolator_kwargs
            Arguments to use when instantiating or calling the extrapolating
            function.
        interpolator
            Interpolator class type to use as interpolating function.
        interpolator_kwargs
            Arguments to use when instantiating the interpolating function.

        Returns
        -------
        :class:`numpy.ndarray`
            Interpolated *RGB* colourspace array.

        Examples
        --------
        >>> LUT = LUT1D(LUT1D.linear_table() ** (1 / 2.2))
        >>> RGB = np.array([0.18, 0.18, 0.18])

        *LUT* applied to the given *RGB* colourspace in the forward direction:

        >>> LUT.apply(RGB)  # doctest: +ELLIPSIS
        array([ 0.4529220...,  0.4529220...,  0.4529220...])

        *LUT* applied to the modified *RGB* colourspace in the inverse
        direction:

        >>> LUT.apply(LUT.apply(RGB), direction='Inverse')
        ... # doctest: +ELLIPSIS
        array([ 0.18...,  0.18...,  0.18...])
        """

        direction = validate_method(
            kwargs.get("direction", "Forward"), ["Forward", "Inverse"]
        )

        interpolator = kwargs.get("interpolator", LinearInterpolator)
        interpolator_kwargs = kwargs.get("interpolator_kwargs", {})
        extrapolator = kwargs.get("extrapolator", Extrapolator)
        extrapolator_kwargs = kwargs.get("extrapolator_kwargs", {})

        LUT = self.invert() if direction == "inverse" else self

        if LUT.is_domain_explicit():
            samples = LUT.domain
        else:
            domain_min, domain_max = LUT.domain
            samples = np.linspace(domain_min, domain_max, LUT.size)

        RGB_interpolator = extrapolator(
            interpolator(samples, LUT.table, **interpolator_kwargs),
            **extrapolator_kwargs,
        )

        return RGB_interpolator(RGB)

    def convert(
        self,
        cls: Type[AbstractLUT],
        force_conversion: Boolean = False,
        **kwargs: Any,
    ) -> AbstractLUT:
        """
        Convert the *LUT* to given ``cls`` class instance.

        Parameters
        ----------
        cls
            *LUT* class instance.
        force_conversion
            Whether to force the conversion as it might be destructive.

        Other Parameters
        ----------------
        interpolator
            Interpolator class type to use as interpolating function.
        interpolator_kwargs
            Arguments to use when instantiating the interpolating function.
        size
            Expected table size in case of an upcast to a :class:`LUT3D` class
            instance.

        Returns
        -------
        :class:`colour.LUT1D` or :class:`colour.LUT3x1D` or \
:class:`colour.LUT3D`
            Converted *LUT* class instance.

        Warnings
        --------
        Some conversions are destructive and raise a :class:`ValueError`
        exception by default.

        Raises
        ------
        ValueError
            If the conversion is destructive.

        Examples
        --------
        >>> LUT = LUT1D()
        >>> print(LUT.convert(LUT1D))
        LUT1D - Unity 10 - Converted 1D to 1D
        -------------------------------------
        <BLANKLINE>
        Dimensions : 1
        Domain     : [ 0.  1.]
        Size       : (10,)
        >>> print(LUT.convert(LUT3x1D))
        LUT3x1D - Unity 10 - Converted 1D to 3x1D
        -----------------------------------------
        <BLANKLINE>
        Dimensions : 2
        Domain     : [[ 0.  0.  0.]
                      [ 1.  1.  1.]]
        Size       : (10, 3)
        >>> print(LUT.convert(LUT3D, force_conversion=True))
        LUT3D - Unity 10 - Converted 1D to 3D
        -------------------------------------
        <BLANKLINE>
        Dimensions : 3
        Domain     : [[ 0.  0.  0.]
                      [ 1.  1.  1.]]
        Size       : (33, 33, 33, 3)
        """

        return LUT_to_LUT(self, cls, force_conversion, **kwargs)

    # ------------------------------------------------------------------------#
    # ---              API Changes and Deprecation Management              ---#
    # ------------------------------------------------------------------------#
    def as_LUT(
        self,
        cls: Type[AbstractLUT],
        force_conversion: Boolean = False,
        **kwargs: Any,
    ) -> AbstractLUT:  # pragma: no cover  # noqa: D102
        # Docstrings are omitted for documentation purposes.
        usage_warning(
            str(
                ObjectRenamed(
                    "LUT1D.as_LUT",
                    "LUT1D.convert",
                )
            )
        )

        return self.convert(cls, force_conversion, **kwargs)


class LUT3x1D(AbstractLUT):
    """
    Define the base class for a 3x1D *LUT*.

    Parameters
    ----------
    table
        Underlying *LUT* table.
    name
        *LUT* name.
    domain
        *LUT* domain, also used to define the instantiation time default table
        domain.
    size
        Size of the instantiation time default table, default to 10.
    comments
        Comments to add to the *LUT*.

    Methods
    -------
    -   :meth:`~colour.LUT3x1D.__init__`
    -   :meth:`~colour.LUT3x1D.is_domain_explicit`
    -   :meth:`~colour.LUT3x1D.linear_table`
    -   :meth:`~colour.LUT3x1D.invert`
    -   :meth:`~colour.LUT3x1D.apply`
    -   :meth:`~colour.LUT3x1D.convert`

    Examples
    --------
    Instantiating a unity LUT with a table with 16x3 elements:

    >>> print(LUT3x1D(size=16))
    LUT3x1D - Unity 16
    ------------------
    <BLANKLINE>
    Dimensions : 2
    Domain     : [[ 0.  0.  0.]
                  [ 1.  1.  1.]]
    Size       : (16, 3)

    Instantiating a LUT using a custom table with 16x3 elements:

    >>> print(LUT3x1D(LUT3x1D.linear_table(16) ** (1 / 2.2)))
    ... # doctest: +ELLIPSIS
    LUT3x1D - ...
    ----------...
    <BLANKLINE>
    Dimensions : 2
    Domain     : [[ 0.  0.  0.]
                  [ 1.  1.  1.]]
    Size       : (16, 3)

    Instantiating a LUT using a custom table with 16x3 elements, custom name,
    custom domain and comments:

    >>> from colour.algebra import spow
    >>> domain = np.array([[-0.1, -0.2, -0.4], [1.5, 3.0, 6.0]])
    >>> print(LUT3x1D(
    ...     spow(LUT3x1D.linear_table(16), 1 / 2.2),
    ...     'My LUT',
    ...     domain,
    ...     comments=['A first comment.', 'A second comment.']))
    LUT3x1D - My LUT
    ----------------
    <BLANKLINE>
    Dimensions : 2
    Domain     : [[-0.1 -0.2 -0.4]
                  [ 1.5  3.   6. ]]
    Size       : (16, 3)
    Comment 01 : A first comment.
    Comment 02 : A second comment.
    """

    def __init__(
        self,
        table: Optional[ArrayLike] = None,
        name: Optional[str] = None,
        domain: Optional[ArrayLike] = None,
        size: Optional[IntegerOrArrayLike] = None,
        comments: Optional[Sequence] = None,
    ):
        domain = cast(
            ArrayLike, optional(domain, np.array([[0, 0, 0], [1, 1, 1]]))
        )
        size = cast(Integer, optional(size, 10))

        super().__init__(table, name, 2, domain, size, comments)

    def _validate_table(self, table: ArrayLike) -> NDArray:
        """
        Validate given table is a 3x1D array.

        Parameters
        ----------
        table
            Table to validate.

        Returns
        -------
        :class:`numpy.ndarray`
            Validated table as a :class:`ndarray` instance.
        """

        table = as_float_array(table)

        attest(len(table.shape) == 2, "The table must be a 2D array!")

        return table

    def _validate_domain(self, domain: ArrayLike) -> NDArray:
        """
        Validate given domain.

        Parameters
        ----------
        domain
            Domain to validate.

        Returns
        -------
        :class:`numpy.ndarray`
            Validated domain as a :class:`ndarray` instance.
        """

        domain = as_float_array(domain)

        attest(len(domain.shape) == 2, "The domain must be a 2D array!")

        attest(
            domain.shape[0] >= 2,
            "The domain row count must be equal or greater than 2!",
        )

        attest(
            domain.shape[1] == 3, "The domain column count must be equal to 3!"
        )

        return domain

    def is_domain_explicit(self) -> Boolean:
        """
        Return whether the *LUT* domain is explicit (or implicit).

        An implicit domain is defined by its shape only::

            [[0 1]
             [0 1]
             [0 1]]

        While an explicit domain defines every single discrete samples::

            [[0.0 0.0 0.0]
             [0.1 0.1 0.1]
             [0.2 0.2 0.2]
             [0.3 0.3 0.3]
             [0.4 0.4 0.4]
             [0.8 0.8 0.8]
             [1.0 1.0 1.0]]

        Returns
        -------
        :class:`bool`
            Is *LUT* domain explicit.

        Examples
        --------
        >>> LUT3x1D().is_domain_explicit()
        False
        >>> samples = np.linspace(0, 1, 10)
        >>> table = domain = tstack([samples, samples, samples])
        >>> LUT3x1D(table, domain=domain).is_domain_explicit()
        True
        """

        return self.domain.shape != (2, 3)

    @staticmethod
    def linear_table(
        size: Optional[IntegerOrArrayLike] = None,
        domain: Optional[ArrayLike] = None,
    ) -> NDArray:
        """
        Return a linear table, the number of output samples :math:`n` is equal
        to ``size * 3`` or ``size[0] + size[1] + size[2]``.

        Parameters
        ----------
        size
            Expected table size, default to 10.
        domain
            Domain of the table.

        Returns
        -------
        :class:`numpy.ndarray`
            Linear table with ``size * 3`` or ``size[0] + size[1] + size[2]``
            samples.

        Warnings
        --------
        If ``size`` is non uniform, the linear table will be padded
        accordingly.

        Examples
        --------
        >>> LUT3x1D.linear_table(
        ...     5, np.array([[-0.1, -0.2, -0.4], [1.5, 3.0, 6.0]]))
        array([[-0.1, -0.2, -0.4],
               [ 0.3,  0.6,  1.2],
               [ 0.7,  1.4,  2.8],
               [ 1.1,  2.2,  4.4],
               [ 1.5,  3. ,  6. ]])
        >>> LUT3x1D.linear_table(
        ...     np.array([5, 3, 2]),
        ...     np.array([[-0.1, -0.2, -0.4], [1.5, 3.0, 6.0]]))
        array([[-0.1, -0.2, -0.4],
               [ 0.3,  1.4,  6. ],
               [ 0.7,  3. ,  nan],
               [ 1.1,  nan,  nan],
               [ 1.5,  nan,  nan]])
        >>> domain = np.array([[-0.1, -0.2, -0.4],
        ...                    [0.3, 1.4, 6.0],
        ...                    [0.7, 3.0, np.nan],
        ...                    [1.1, np.nan, np.nan],
        ...                    [1.5, np.nan, np.nan]])
        >>> LUT3x1D.linear_table(domain=domain)
        array([[-0.1, -0.2, -0.4],
               [ 0.3,  1.4,  6. ],
               [ 0.7,  3. ,  nan],
               [ 1.1,  nan,  nan],
               [ 1.5,  nan,  nan]])
        """

        size = cast(Integer, optional(size, 10))
        domain = as_float_array(
            cast(ArrayLike, optional(domain, np.array([[0, 0, 0], [1, 1, 1]])))
        )

        if domain.shape != (2, 3):
            return domain
        else:
            if is_numeric(size):
                size_array = np.tile(size, 3)
            else:
                size_array = as_int_array(size)

            R, G, B = tsplit(domain)

            samples = [
                np.linspace(a[0], a[1], size_array[i])
                for i, a in enumerate([R, G, B])
            ]

            if not len(np.unique(size_array)) == 1:
                runtime_warning(
                    "Table is non uniform, axis will be "
                    'padded with "NaNs" accordingly!'
                )

                samples = [
                    np.pad(
                        axis,
                        (0, np.max(size_array) - len(axis)),
                        mode="constant",
                        constant_values=np.nan,
                    )
                    for axis in samples
                ]

            return tstack(samples)

    def invert(self, **kwargs: Any) -> LUT3x1D:
        """
        Compute and returns an inverse copy of the *LUT*.

        Other Parameters
        ----------------
        kwargs
            Keywords arguments, only given for signature compatibility with
            the :meth:`AbstractLUT.invert` method.

        Returns
        -------
        :class:`colour.LUT3x1D`
            Inverse *LUT* class instance.

        Examples
        --------
        >>> LUT = LUT3x1D(LUT3x1D.linear_table() ** (1 / 2.2))
        >>> print(LUT.table)
        [[ 0.          0.          0.        ]
         [ 0.36834383  0.36834383  0.36834383]
         [ 0.50476034  0.50476034  0.50476034]
         [ 0.60691337  0.60691337  0.60691337]
         [ 0.69169882  0.69169882  0.69169882]
         [ 0.76553851  0.76553851  0.76553851]
         [ 0.83168433  0.83168433  0.83168433]
         [ 0.89204934  0.89204934  0.89204934]
         [ 0.94787016  0.94787016  0.94787016]
         [ 1.          1.          1.        ]]
        >>> print(LUT.invert())  # doctest: +ELLIPSIS
        LUT3x1D - ... - Inverse
        ----------...----------
        <BLANKLINE>
        Dimensions : 2
        Domain     : [[ 0.       ...  0.       ...  0.       ...]
                      [ 0.3683438...  0.3683438...  0.3683438...]
                      [ 0.5047603...  0.5047603...  0.5047603...]
                      [ 0.6069133...  0.6069133...  0.6069133...]
                      [ 0.6916988...  0.6916988...  0.6916988...]
                      [ 0.7655385...  0.7655385...  0.7655385...]
                      [ 0.8316843...  0.8316843...  0.8316843...]
                      [ 0.8920493...  0.8920493...  0.8920493...]
                      [ 0.9478701...  0.9478701...  0.9478701...]
                      [ 1.       ...  1.       ...  1.       ...]]
        Size       : (10, 3)
        >>> print(LUT.invert().table)  # doctest: +ELLIPSIS
        [[ 0.       ...  0.       ...  0.       ...]
         [ 0.1111111...  0.1111111...  0.1111111...]
         [ 0.2222222...  0.2222222...  0.2222222...]
         [ 0.3333333...  0.3333333...  0.3333333...]
         [ 0.4444444...  0.4444444...  0.4444444...]
         [ 0.5555555...  0.5555555...  0.5555555...]
         [ 0.6666666...  0.6666666...  0.6666666...]
         [ 0.7777777...  0.7777777...  0.7777777...]
         [ 0.8888888...  0.8888888...  0.8888888...]
         [ 1.       ...  1.       ...  1.       ...]]
        """

        size = self.table.size // 3
        if self.is_domain_explicit():
            domain = [
                axes[: (~np.isnan(axes)).cumsum().argmax() + 1]
                for axes in np.transpose(self.domain)
            ]
        else:
            domain_min, domain_max = self.domain
            domain = [
                np.linspace(domain_min[i], domain_max[i], size)
                for i in range(3)
            ]

        LUT_i = LUT3x1D(
            table=tstack(domain),
            name=f"{self.name} - Inverse",
            domain=self.table,
        )

        return LUT_i

    def apply(self, RGB: ArrayLike, **kwargs: Any) -> NDArray:
        """
        Apply the *LUT* to given *RGB* colourspace array using given method.

        Parameters
        ----------
        RGB
            *RGB* colourspace array to apply the *LUT* onto.

        Other Parameters
        ----------------
        direction
            Whether the *LUT* should be applied in the forward or inverse
            direction.
        extrapolator
            Extrapolator class type or object to use as extrapolating function.
        extrapolator_kwargs
            Arguments to use when instantiating or calling the extrapolating
            function.
        interpolator
            Interpolator class type to use as interpolating function.
        interpolator_kwargs
            Arguments to use when instantiating the interpolating function.

        Returns
        -------
        :class:`numpy.ndarray`
            Interpolated *RGB* colourspace array.

        Examples
        --------
        >>> LUT = LUT3x1D(LUT3x1D.linear_table() ** (1 / 2.2))
        >>> RGB = np.array([0.18, 0.18, 0.18])
        >>> LUT.apply(RGB)  # doctest: +ELLIPSIS
        array([ 0.4529220...,  0.4529220...,  0.4529220...])
        >>> LUT.apply(LUT.apply(RGB), direction='Inverse')
        ... # doctest: +ELLIPSIS
        array([ 0.18...,  0.18...,  0.18...])
        >>> from colour.algebra import spow
        >>> domain = np.array([[-0.1, -0.2, -0.4], [1.5, 3.0, 6.0]])
        >>> table = spow(LUT3x1D.linear_table(domain=domain), 1 / 2.2)
        >>> LUT = LUT3x1D(table, domain=domain)
        >>> RGB = np.array([0.18, 0.18, 0.18])
        >>> LUT.apply(RGB)  # doctest: +ELLIPSIS
        array([ 0.4423903...,  0.4503801...,  0.3581625...])
        >>> domain = np.array([[-0.1, -0.2, -0.4],
        ...                    [0.3, 1.4, 6.0],
        ...                    [0.7, 3.0, np.nan],
        ...                    [1.1, np.nan, np.nan],
        ...                    [1.5, np.nan, np.nan]])
        >>> table = spow(LUT3x1D.linear_table(domain=domain), 1 / 2.2)
        >>> LUT = LUT3x1D(table, domain=domain)
        >>> RGB = np.array([0.18, 0.18, 0.18])
        >>> LUT.apply(RGB)  # doctest: +ELLIPSIS
        array([ 0.2996370..., -0.0901332..., -0.3949770...])
        """

        direction = validate_method(
            kwargs.get("direction", "Forward"), ["Forward", "Inverse"]
        )

        interpolator = kwargs.get("interpolator", LinearInterpolator)
        interpolator_kwargs = kwargs.get("interpolator_kwargs", {})
        extrapolator = kwargs.get("extrapolator", Extrapolator)
        extrapolator_kwargs = kwargs.get("extrapolator_kwargs", {})

        R, G, B = tsplit(RGB)

        LUT = self.invert() if direction == "inverse" else self

        size = LUT.table.size // 3
        if LUT.is_domain_explicit():
            samples = [
                axes[: (~np.isnan(axes)).cumsum().argmax() + 1]
                for axes in np.transpose(LUT.domain)
            ]
            R_t, G_t, B_t = (
                axes[: len(samples[i])]
                for i, axes in enumerate(np.transpose(LUT.table))
            )
        else:
            domain_min, domain_max = LUT.domain
            samples = [
                np.linspace(domain_min[i], domain_max[i], size)
                for i in range(3)
            ]
            R_t, G_t, B_t = tsplit(LUT.table)

        s_R, s_G, s_B = samples

        RGB_i = [
            extrapolator(
                interpolator(a[0], a[1], **interpolator_kwargs),
                **extrapolator_kwargs,
            )(a[2])
            for a in zip((s_R, s_G, s_B), (R_t, G_t, B_t), (R, G, B))
        ]

        return tstack(RGB_i)

    def convert(
        self,
        cls: Type[AbstractLUT],
        force_conversion: Boolean = False,
        **kwargs: Any,
    ) -> AbstractLUT:
        """
        Convert the *LUT* to given ``cls`` class instance.

        Parameters
        ----------
        cls
            *LUT* class instance.
        force_conversion
            Whether to force the conversion as it might be destructive.

        Other Parameters
        ----------------
        interpolator
            Interpolator class type to use as interpolating function.
        interpolator_kwargs
            Arguments to use when instantiating the interpolating function.
        size
            Expected table size in case of an upcast to a :class:`LUT3D` class
            instance.

        Returns
        -------
        :class:`colour.LUT1D` or :class:`colour.LUT3x1D` or \
:class:`colour.LUT3D`
            Converted *LUT* class instance.

        Warnings
        --------
        Some conversions are destructive and raise a :class:`ValueError`
        exception by default.

        Raises
        ------
        ValueError
            If the conversion is destructive.

        Examples
        --------
        >>> LUT = LUT3x1D()
        >>> print(LUT.convert(LUT1D, force_conversion=True))
        LUT1D - Unity 10 - Converted 3x1D to 1D
        ---------------------------------------
        <BLANKLINE>
        Dimensions : 1
        Domain     : [ 0.  1.]
        Size       : (10,)
        >>> print(LUT.convert(LUT3x1D))
        LUT3x1D - Unity 10 - Converted 3x1D to 3x1D
        -------------------------------------------
        <BLANKLINE>
        Dimensions : 2
        Domain     : [[ 0.  0.  0.]
                      [ 1.  1.  1.]]
        Size       : (10, 3)
        >>> print(LUT.convert(LUT3D, force_conversion=True))
        LUT3D - Unity 10 - Converted 3x1D to 3D
        ---------------------------------------
        <BLANKLINE>
        Dimensions : 3
        Domain     : [[ 0.  0.  0.]
                      [ 1.  1.  1.]]
        Size       : (33, 33, 33, 3)
        """

        return LUT_to_LUT(self, cls, force_conversion, **kwargs)

    # ------------------------------------------------------------------------#
    # ---              API Changes and Deprecation Management              ---#
    # ------------------------------------------------------------------------#
    def as_LUT(
        self,
        cls: Type[AbstractLUT],
        force_conversion: Boolean = False,
        **kwargs: Any,
    ) -> AbstractLUT:  # pragma: no cover  # noqa: D102
        # Docstrings are omitted for documentation purposes.
        usage_warning(
            str(
                ObjectRenamed(
                    "LUT3x1D.as_LUT",
                    "LUT3x1D.convert",
                )
            )
        )

        return self.convert(cls, force_conversion, **kwargs)


class LUT3D(AbstractLUT):
    """
    Define the base class for a 3D *LUT*.

    Parameters
    ----------
    table
        Underlying *LUT* table.
    name
        *LUT* name.
    domain
        *LUT* domain, also used to define the instantiation time default table
        domain.
    size
        Size of the instantiation time default table, default to 33.
    comments
        Comments to add to the *LUT*.

    Methods
    -------
    -   :meth:`~colour.LUT3D.__init__`
    -   :meth:`~colour.LUT3D.is_domain_explicit`
    -   :meth:`~colour.LUT3D.linear_table`
    -   :meth:`~colour.LUT3D.invert`
    -   :meth:`~colour.LUT3D.apply`
    -   :meth:`~colour.LUT3D.convert`

    Examples
    --------
    Instantiating a unity LUT with a table with 16x16x16x3 elements:

    >>> print(LUT3D(size=16))
    LUT3D - Unity 16
    ----------------
    <BLANKLINE>
    Dimensions : 3
    Domain     : [[ 0.  0.  0.]
                  [ 1.  1.  1.]]
    Size       : (16, 16, 16, 3)

    Instantiating a LUT using a custom table with 16x16x16x3 elements:

    >>> print(LUT3D(LUT3D.linear_table(16) ** (1 / 2.2)))  # doctest: +ELLIPSIS
    LUT3D - ...
    --------...
    <BLANKLINE>
    Dimensions : 3
    Domain     : [[ 0.  0.  0.]
                  [ 1.  1.  1.]]
    Size       : (16, 16, 16, 3)

    Instantiating a LUT using a custom table with 16x16x16x3 elements, custom
    name, custom domain and comments:

    >>> from colour.algebra import spow
    >>> domain = np.array([[-0.1, -0.2, -0.4], [1.5, 3.0, 6.0]])
    >>> print(LUT3D(
    ...     spow(LUT3D.linear_table(16), 1 / 2.2),
    ...     'My LUT',
    ...     domain,
    ...     comments=['A first comment.', 'A second comment.']))
    LUT3D - My LUT
    --------------
    <BLANKLINE>
    Dimensions : 3
    Domain     : [[-0.1 -0.2 -0.4]
                  [ 1.5  3.   6. ]]
    Size       : (16, 16, 16, 3)
    Comment 01 : A first comment.
    Comment 02 : A second comment.
    """

    def __init__(
        self,
        table: Optional[ArrayLike] = None,
        name: Optional[str] = None,
        domain: Optional[ArrayLike] = None,
        size: Optional[IntegerOrArrayLike] = None,
        comments: Optional[Sequence] = None,
    ):
        domain = cast(
            ArrayLike, optional(domain, np.array([[0, 0, 0], [1, 1, 1]]))
        )
        size = cast(Integer, optional(size, 33))

        super().__init__(table, name, 3, domain, size, comments)

    def _validate_table(self, table: ArrayLike) -> NDArray:
        """
        Validate given table is a 4D array and that its dimensions are equal.

        Parameters
        ----------
        table
            Table to validate.

        Returns
        -------
        :class:`numpy.ndarray`
            Validated table as a :class:`ndarray` instance.
        """

        table = as_float_array(table)

        attest(len(table.shape) == 4, "The table must be a 4D array!")

        return table

    def _validate_domain(self, domain: ArrayLike) -> NDArray:
        """
        Validate given domain.

        Parameters
        ----------
        domain
            Domain to validate.

        Returns
        -------
        :class:`numpy.ndarray`
            Validated domain as a :class:`ndarray` instance.

        Notes
        -----
        -   A :class:`LUT3D` class instance must use an implicit domain.
        """

        domain = as_float_array(domain)

        attest(len(domain.shape) == 2, "The domain must be a 2D array!")

        attest(
            domain.shape[0] >= 2,
            "The domain row count must be equal or greater than 2!",
        )

        attest(
            domain.shape[1] == 3, "The domain column count must be equal to 3!"
        )

        return domain

    def is_domain_explicit(self) -> Boolean:
        """
        Return whether the *LUT* domain is explicit (or implicit).

        An implicit domain is defined by its shape only::

            [[0 0 0]
             [1 1 1]]

        While an explicit domain defines every single discrete samples::

            [[0.0 0.0 0.0]
             [0.1 0.1 0.1]
             [0.2 0.2 0.2]
             [0.3 0.3 0.3]
             [0.4 0.4 0.4]
             [0.8 0.8 0.8]
             [1.0 1.0 1.0]]

        Returns
        -------
        :class:`bool`
            Is *LUT* domain explicit.

        Examples
        --------
        >>> LUT3D().is_domain_explicit()
        False
        >>> domain = np.array([[-0.1, -0.2, -0.4],
        ...                    [0.7, 1.4, 6.0],
        ...                    [1.5, 3.0, np.nan]])
        >>> LUT3D(domain=domain).is_domain_explicit()
        True
        """

        return self.domain.shape != (2, 3)

    @staticmethod
    def linear_table(
        size: Optional[IntegerOrArrayLike] = None,
        domain: Optional[ArrayLike] = None,
    ) -> NDArray:
        """
        Return a linear table, the number of output samples :math:`n` is equal
        to ``size**3 * 3`` or ``size[0] * size[1] * size[2] * 3``.

        Parameters
        ----------
        size
            Expected table size, default to 33.
        domain
            Domain of the table.

        Returns
        -------
        :class:`numpy.ndarray`
            Linear table with ``size**3 * 3`` or
            ``size[0] * size[1] * size[2] * 3`` samples.

        Examples
        --------
        >>> LUT3D.linear_table(
        ...     3, np.array([[-0.1, -0.2, -0.4], [1.5, 3.0, 6.0]]))
        array([[[[-0.1, -0.2, -0.4],
                 [-0.1, -0.2,  2.8],
                 [-0.1, -0.2,  6. ]],
        <BLANKLINE>
                [[-0.1,  1.4, -0.4],
                 [-0.1,  1.4,  2.8],
                 [-0.1,  1.4,  6. ]],
        <BLANKLINE>
                [[-0.1,  3. , -0.4],
                 [-0.1,  3. ,  2.8],
                 [-0.1,  3. ,  6. ]]],
        <BLANKLINE>
        <BLANKLINE>
               [[[ 0.7, -0.2, -0.4],
                 [ 0.7, -0.2,  2.8],
                 [ 0.7, -0.2,  6. ]],
        <BLANKLINE>
                [[ 0.7,  1.4, -0.4],
                 [ 0.7,  1.4,  2.8],
                 [ 0.7,  1.4,  6. ]],
        <BLANKLINE>
                [[ 0.7,  3. , -0.4],
                 [ 0.7,  3. ,  2.8],
                 [ 0.7,  3. ,  6. ]]],
        <BLANKLINE>
        <BLANKLINE>
               [[[ 1.5, -0.2, -0.4],
                 [ 1.5, -0.2,  2.8],
                 [ 1.5, -0.2,  6. ]],
        <BLANKLINE>
                [[ 1.5,  1.4, -0.4],
                 [ 1.5,  1.4,  2.8],
                 [ 1.5,  1.4,  6. ]],
        <BLANKLINE>
                [[ 1.5,  3. , -0.4],
                 [ 1.5,  3. ,  2.8],
                 [ 1.5,  3. ,  6. ]]]])
        >>> LUT3D.linear_table(
        ...     np.array([3, 3, 2]),
        ...     np.array([[-0.1, -0.2, -0.4], [1.5, 3.0, 6.0]]))
        array([[[[-0.1, -0.2, -0.4],
                 [-0.1, -0.2,  6. ]],
        <BLANKLINE>
                [[-0.1,  1.4, -0.4],
                 [-0.1,  1.4,  6. ]],
        <BLANKLINE>
                [[-0.1,  3. , -0.4],
                 [-0.1,  3. ,  6. ]]],
        <BLANKLINE>
        <BLANKLINE>
               [[[ 0.7, -0.2, -0.4],
                 [ 0.7, -0.2,  6. ]],
        <BLANKLINE>
                [[ 0.7,  1.4, -0.4],
                 [ 0.7,  1.4,  6. ]],
        <BLANKLINE>
                [[ 0.7,  3. , -0.4],
                 [ 0.7,  3. ,  6. ]]],
        <BLANKLINE>
        <BLANKLINE>
               [[[ 1.5, -0.2, -0.4],
                 [ 1.5, -0.2,  6. ]],
        <BLANKLINE>
                [[ 1.5,  1.4, -0.4],
                 [ 1.5,  1.4,  6. ]],
        <BLANKLINE>
                [[ 1.5,  3. , -0.4],
                 [ 1.5,  3. ,  6. ]]]])
        >>> domain = np.array([[-0.1, -0.2, -0.4],
        ...                    [0.7, 1.4, 6.0],
        ...                    [1.5, 3.0, np.nan]])
        >>> LUT3D.linear_table(domain=domain)
        array([[[[-0.1, -0.2, -0.4],
                 [-0.1, -0.2,  6. ]],
        <BLANKLINE>
                [[-0.1,  1.4, -0.4],
                 [-0.1,  1.4,  6. ]],
        <BLANKLINE>
                [[-0.1,  3. , -0.4],
                 [-0.1,  3. ,  6. ]]],
        <BLANKLINE>
        <BLANKLINE>
               [[[ 0.7, -0.2, -0.4],
                 [ 0.7, -0.2,  6. ]],
        <BLANKLINE>
                [[ 0.7,  1.4, -0.4],
                 [ 0.7,  1.4,  6. ]],
        <BLANKLINE>
                [[ 0.7,  3. , -0.4],
                 [ 0.7,  3. ,  6. ]]],
        <BLANKLINE>
        <BLANKLINE>
               [[[ 1.5, -0.2, -0.4],
                 [ 1.5, -0.2,  6. ]],
        <BLANKLINE>
                [[ 1.5,  1.4, -0.4],
                 [ 1.5,  1.4,  6. ]],
        <BLANKLINE>
                [[ 1.5,  3. , -0.4],
                 [ 1.5,  3. ,  6. ]]]])
        """

        size = cast(Integer, optional(size, 33))
        domain = as_float_array(
            cast(ArrayLike, optional(domain, np.array([[0, 0, 0], [1, 1, 1]])))
        )

        if domain.shape != (2, 3):
            samples = list(
                np.flip(
                    [
                        axes[: (~np.isnan(axes)).cumsum().argmax() + 1]
                        for axes in np.transpose(domain)
                    ],
                    -1,
                )
            )
            size_array = as_int_array([len(axes) for axes in samples])
        else:
            if is_numeric(size):
                size_array = np.tile(size, 3)
            else:
                size_array = as_int_array(size)

            R, G, B = tsplit(domain)

            size_array = np.flip(size_array, -1)
            samples = [
                np.linspace(a[0], a[1], size_array[i])
                for i, a in enumerate([B, G, R])
            ]

        table = np.flip(
            np.transpose(np.meshgrid(*samples, indexing="ij")).reshape(
                np.hstack([np.flip(size_array, -1), 3])
            ),
            -1,
        )

        return table

    @required("Scikit-Learn")
    def invert(self, **kwargs: Any) -> LUT3D:
        """
        Compute and returns an inverse copy of the *LUT*.

        Other Parameters
        ----------------
        extrapolate
            Whether to extrapolate the *LUT* when computing its inverse.
            Extrapolation is performed by reflecting the *LUT* cube along its 8
            faces. Note that the domain is extended beyond [0, 1], thus the
            *LUT* might not be handled properly in other software.
        interpolator
            Interpolator class type or object to use as interpolating function.
        query_size
            Number of points to query in the KDTree, their mean is computed,
            resulting in a smoother result.
        size
            Size of the inverse *LUT*. With the given implementation, it is
            good practise to double the size of the inverse *LUT* to provide a
            smoother result. If ``size`` is not given,
            :math:`2^{\\sqrt{size_{LUT}} + 1} + 1` will be used instead.

        Returns
        -------
        :class:`colour.LUT3D`
            Inverse *LUT* class instance.

        Examples
        --------
        >>> LUT = LUT3D()
        >>> print(LUT)
        LUT3D - Unity 33
        ----------------
        <BLANKLINE>
        Dimensions : 3
        Domain     : [[ 0.  0.  0.]
                      [ 1.  1.  1.]]
        Size       : (33, 33, 33, 3)
        >>> print(LUT.invert())
        LUT3D - Unity 33 - Inverse
        --------------------------
        <BLANKLINE>
        Dimensions : 3
        Domain     : [[ 0.  0.  0.]
                      [ 1.  1.  1.]]
        Size       : (108, 108, 108, 3)
        """

        # TODO: Drop "sklearn" requirement whenever "Scipy" 1.7 can be
        # defined as the minimal version.
        from sklearn.neighbors import KDTree

        interpolator = kwargs.get(
            "interpolator", table_interpolation_trilinear
        )
        extrapolate = kwargs.get("extrapolate", False)
        query_size = kwargs.get("query_size", 3)

        LUT = self.copy()
        source_size = LUT.size
        target_size = kwargs.get(
            "size", (as_int(2 ** (np.sqrt(source_size) + 1) + 1))
        )

        if target_size > 129:
            usage_warning("LUT3D inverse computation time could be excessive!")

        if extrapolate:
            LUT.table = np.pad(
                LUT.table,
                [(1, 1), (1, 1), (1, 1), (0, 0)],
                "reflect",
                reflect_type="odd",
            )

            LUT.domain[0] -= 1 / (source_size - 1)
            LUT.domain[1] += 1 / (source_size - 1)

        # "LUT_t" is an intermediate LUT with a size equal to that of the
        # final inverse LUT which is usually larger than the input LUT.
        # The intent is to smooth the inverse LUT's table by increasing the
        # resolution of the KDTree.
        LUT_t = LUT3D(size=target_size, domain=LUT.domain)
        table = np.reshape(LUT_t.table, (-1, 3))
        LUT_t.table = LUT.apply(LUT_t.table, interpolator=interpolator)

        tree = KDTree(np.reshape(LUT_t.table, (-1, 3)))

        # "LUT_q" stores the indexes of the KDTree query, i.e. the closest
        # entry of "LUT_t" for any searched table sample.
        LUT_q = LUT3D(size=target_size, domain=LUT.domain)
        query = tree.query(table, query_size)[-1]
        if query_size == 1:
            LUT_q.table = table[query].reshape(
                [target_size, target_size, target_size, 3]
            )
        else:
            LUT_q.table = np.mean(table[query], axis=-2).reshape(
                [target_size, target_size, target_size, 3]
            )

        # "LUT_i" is the final inverse LUT generated by applying "LUT_q" on
        # an identity LUT at the target size.
        LUT_i = LUT3D(size=target_size, domain=LUT.domain)
        LUT_i.table = LUT_q.apply(LUT_i.table, interpolator=interpolator)

        LUT_i.name = f"{self.name} - Inverse"

        return LUT_i

    def apply(self, RGB: ArrayLike, **kwargs: Any) -> NDArray:
        """
        Apply the *LUT* to given *RGB* colourspace array using given method.

        Parameters
        ----------
        RGB
            *RGB* colourspace array to apply the *LUT* onto.

        Other Parameters
        ----------------
        direction
            Whether the *LUT* should be applied in the forward or inverse
            direction.
        extrapolate
            Whether to extrapolate the *LUT* when computing its inverse.
            Extrapolation is performed by reflecting the *LUT* cube along its 8
            faces.
        interpolator
            Interpolator object to use as interpolating function.
        interpolator_kwargs
            Arguments to use when calling the interpolating function.
        query_size
            Number of points to query in the KDTree, their mean is computed,
            resulting in a smoother result.
        size
            Size of the inverse *LUT*. With the given implementation, it is
            good practise to double the size of the inverse *LUT* to provide a
            smoother result. If ``size`` is not given,
            :math:`2^{\\sqrt{size_{LUT}} + 1} + 1` will be used instead.

        Returns
        -------
        :class:`numpy.ndarray`
            Interpolated *RGB* colourspace array.

        Examples
        --------
        >>> LUT = LUT3D(LUT3D.linear_table() ** (1 / 2.2))
        >>> RGB = np.array([0.18, 0.18, 0.18])
        >>> LUT.apply(RGB)  # doctest: +ELLIPSIS
        array([ 0.4583277...,  0.4583277...,  0.4583277...])
        >>> LUT.apply(LUT.apply(RGB), direction='Inverse')
        ... # doctest: +ELLIPSIS
        array([ 0.1781995...,  0.1809414...,  0.1809513...])
        >>> from colour.algebra import spow
        >>> domain = np.array([[-0.1, -0.2, -0.4],
        ...                    [0.3, 1.4, 6.0],
        ...                    [0.7, 3.0, np.nan],
        ...                    [1.1, np.nan, np.nan],
        ...                    [1.5, np.nan, np.nan]])
        >>> table = spow(LUT3D.linear_table(domain=domain), 1 / 2.2)
        >>> LUT = LUT3D(table, domain=domain)
        >>> RGB = np.array([0.18, 0.18, 0.18])
        >>> LUT.apply(RGB)  # doctest: +ELLIPSIS
        array([ 0.2996370..., -0.0901332..., -0.3949770...])
        """

        direction = validate_method(
            kwargs.get("direction", "Forward"), ["Forward", "Inverse"]
        )

        interpolator = kwargs.get(
            "interpolator", table_interpolation_trilinear
        )
        interpolator_kwargs = kwargs.get("interpolator_kwargs", {})

        R, G, B = tsplit(RGB)

        settings = {"interpolator": interpolator}
        settings.update(**kwargs)
        LUT = self.invert(**settings) if direction == "inverse" else self

        if LUT.is_domain_explicit():
            domain_min = LUT.domain[0, ...]
            domain_max = [
                axes[: (~np.isnan(axes)).cumsum().argmax() + 1][-1]
                for axes in np.transpose(LUT.domain)
            ]
            usage_warning(
                f'"LUT" was defined with an explicit domain but requires an '
                f"implicit domain to be applied. The following domain will be "
                f"used: {np.vstack([domain_min, domain_max])}"
            )
        else:
            domain_min, domain_max = LUT.domain

        RGB_l = [
            linear_conversion(j, (domain_min[i], domain_max[i]), (0, 1))
            for i, j in enumerate((R, G, B))
        ]

        RGB_i = interpolator(tstack(RGB_l), LUT.table, **interpolator_kwargs)

        return RGB_i

    def convert(
        self,
        cls: Type[AbstractLUT],
        force_conversion: Boolean = False,
        **kwargs: Any,
    ) -> AbstractLUT:
        """
        Convert the *LUT* to given ``cls`` class instance.

        Parameters
        ----------
        cls
            *LUT* class instance.
        force_conversion
            Whether to force the conversion as it might be destructive.

        Other Parameters
        ----------------
        interpolator
            Interpolator class type to use as interpolating function.
        interpolator_kwargs
            Arguments to use when instantiating the interpolating function.
        size
            Expected table size in case of a downcast from a :class:`LUT3D`
            class instance.

        Returns
        -------
        :class:`colour.LUT1D` or :class:`colour.LUT3x1D` or \
:class:`colour.LUT3D`
            Converted *LUT* class instance.

        Warnings
        --------
        Some conversions are destructive and raise a :class:`ValueError`
        exception by default.

        Raises
        ------
        ValueError
            If the conversion is destructive.

        Examples
        --------
        >>> LUT = LUT3D()
        >>> print(LUT.convert(LUT1D, force_conversion=True))
        LUT1D - Unity 33 - Converted 3D to 1D
        -------------------------------------
        <BLANKLINE>
        Dimensions : 1
        Domain     : [ 0.  1.]
        Size       : (10,)
        >>> print(LUT.convert(LUT3x1D, force_conversion=True))
        LUT3x1D - Unity 33 - Converted 3D to 3x1D
        -----------------------------------------
        <BLANKLINE>
        Dimensions : 2
        Domain     : [[ 0.  0.  0.]
                      [ 1.  1.  1.]]
        Size       : (10, 3)
        >>> print(LUT.convert(LUT3D))
        LUT3D - Unity 33 - Converted 3D to 3D
        -------------------------------------
        <BLANKLINE>
        Dimensions : 3
        Domain     : [[ 0.  0.  0.]
                      [ 1.  1.  1.]]
        Size       : (33, 33, 33, 3)
        """

        return LUT_to_LUT(self, cls, force_conversion, **kwargs)

    # ------------------------------------------------------------------------#
    # ---              API Changes and Deprecation Management              ---#
    # ------------------------------------------------------------------------#
    def as_LUT(
        self,
        cls: Type[AbstractLUT],
        force_conversion: Boolean = False,
        **kwargs: Any,
    ) -> AbstractLUT:  # pragma: no cover  # noqa: D102
        # Docstrings are omitted for documentation purposes.
        usage_warning(
            str(
                ObjectRenamed(
                    "LUT3D.as_LUT",
                    "LUT3D.convert",
                )
            )
        )

        return self.convert(cls, force_conversion, **kwargs)


def LUT_to_LUT(
    LUT,
    cls: Type[AbstractLUT],
    force_conversion: Boolean = False,
    **kwargs: Any,
) -> AbstractLUT:
    """
    Convert given *LUT* to given ``cls`` class instance.

    Parameters
    ----------
    cls
        *LUT* class instance.
    force_conversion
        Whether to force the conversion if destructive.

    Other Parameters
    ----------------
    channel_weights
        Channel weights in case of a downcast from a :class:`LUT3x1D` or
        :class:`LUT3D` class instance.
    interpolator
        Interpolator class type to use as interpolating function.
    interpolator_kwargs
        Arguments to use when instantiating the interpolating function.
    size
        Expected table size in case of an upcast to or a downcast from a
        :class:`LUT3D` class instance.

    Returns
    -------
    :class:`colour.LUT1D` or :class:`colour.LUT3x1D` or :class:`colour.LUT3D`
        Converted *LUT* class instance.

    Warnings
    --------
    Some conversions are destructive and raise a :class:`ValueError` exception
    by default.

    Raises
    ------
    ValueError
        If the conversion is destructive.

    Examples
    --------
    >>> print(LUT_to_LUT(LUT1D(), LUT3D, force_conversion=True))
    LUT3D - Unity 10 - Converted 1D to 3D
    -------------------------------------
    <BLANKLINE>
    Dimensions : 3
    Domain     : [[ 0.  0.  0.]
                  [ 1.  1.  1.]]
    Size       : (33, 33, 33, 3)
    >>> print(LUT_to_LUT(LUT3x1D(), LUT1D, force_conversion=True))
    LUT1D - Unity 10 - Converted 3x1D to 1D
    ---------------------------------------
    <BLANKLINE>
    Dimensions : 1
    Domain     : [ 0.  1.]
    Size       : (10,)
    >>> print(LUT_to_LUT(LUT3D(), LUT1D, force_conversion=True))
    LUT1D - Unity 33 - Converted 3D to 1D
    -------------------------------------
    <BLANKLINE>
    Dimensions : 1
    Domain     : [ 0.  1.]
    Size       : (10,)
    """

    ranks = {LUT1D: 1, LUT3x1D: 2, LUT3D: 3}
    path = (ranks[LUT.__class__], ranks[cls])
    path_verbose = [
        f"{element}D" if element != 2 else "3x1D" for element in path
    ]
    if path in ((1, 3), (2, 1), (2, 3), (3, 1), (3, 2)):
        if not force_conversion:
            raise ValueError(
                f'Conversion of a "LUT" {path_verbose[0]} to a "LUT" '
                f"{path_verbose[1]} is destructive, please use the "
                f'"force_conversion" argument to proceed!'
            )

    suffix = f" - Converted {path_verbose[0]} to {path_verbose[1]}"
    name = f"{LUT.name}{suffix}"

    # Same dimension conversion, returning a copy.
    if len(set(path)) == 1:
        LUT = LUT.copy()
        LUT.name = name
    else:
        size = kwargs.get("size", 33 if cls is LUT3D else 10)
        if "size" in kwargs:
            del kwargs["size"]

        channel_weights = as_float_array(
            kwargs.get("channel_weights", full(3, 1 / 3))
        )
        if "channel_weights" in kwargs:
            del kwargs["channel_weights"]

        if isinstance(LUT, LUT1D):
            if cls is LUT3x1D:
                domain = tstack([LUT.domain, LUT.domain, LUT.domain])
                table = tstack([LUT.table, LUT.table, LUT.table])
            elif cls is LUT3D:
                domain = tstack([LUT.domain, LUT.domain, LUT.domain])
                table = LUT3D.linear_table(size, domain)
                table = LUT.apply(table, **kwargs)
        elif isinstance(LUT, LUT3x1D):
            if cls is LUT1D:
                domain = np.sum(LUT.domain * channel_weights, axis=-1)
                table = np.sum(LUT.table * channel_weights, axis=-1)
            elif cls is LUT3D:
                domain = LUT.domain
                table = LUT3D.linear_table(size, domain)
                table = LUT.apply(table, **kwargs)
        elif isinstance(LUT, LUT3D):
            if cls is LUT1D:
                domain = np.sum(LUT.domain * channel_weights, axis=-1)
                table = LUT1D.linear_table(size, domain)
                table = LUT.apply(tstack([table, table, table]), **kwargs)
                table = np.sum(table * channel_weights, axis=-1)
            elif cls is LUT3x1D:
                domain = LUT.domain
                table = LUT3x1D.linear_table(size, domain)
                table = LUT.apply(table, **kwargs)

        LUT = cls(
            table=table,
            name=name,
            domain=domain,
            size=table.shape[0],
            comments=LUT.comments,
        )

    return LUT



"""Defines the unit tests for the :mod:`colour.contrast.barten1999` module."""

import numpy as np
import unittest
from itertools import permutations

from colour.contrast import (
    optical_MTF_Barten1999,
    pupil_diameter_Barten1999,
    sigma_Barten1999,
    retinal_illuminance_Barten1999,
    maximum_angular_size_Barten1999,
    contrast_sensitivity_function_Barten1999,
)
from colour.utilities import ignore_numpy_errors

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestOpticalMTFBarten1999",
    "TestPupilDiameterBarten1999",
    "TestSigmaBarten1999",
    "TestRetinalIlluminanceBarten1999",
    "TestMaximumAngularSizeBarten1999",
    "TestContrastSensitivityFunctionBarten1999",
]


class TestOpticalMTFBarten1999(unittest.TestCase):
    """
    Define :func:`colour.contrast.barten1999.optical_MTF_Barten1999`
    definition unit tests methods.
    """

    def test_optical_MTF_Barten1999(self):
        """
        Test :func:`colour.contrast.barten1999.optical_MTF_Barten1999`
        definition.
        """

        np.testing.assert_almost_equal(
            optical_MTF_Barten1999(4, 0.01), 0.968910791191297, decimal=7
        )

        np.testing.assert_almost_equal(
            optical_MTF_Barten1999(8, 0.01), 0.881323136669471, decimal=7
        )

        np.testing.assert_almost_equal(
            optical_MTF_Barten1999(4, 0.05), 0.454040738727245, decimal=7
        )

    def test_n_dimensional_optical_MTF_Barten1999(self):
        """
        Test :func:`colour.contrast.barten1999.optical_MTF_Barten1999`
        definition n-dimensional support.
        """

        u = np.array([4, 8, 12])
        sigma = np.array([0.01, 0.05, 0.1])
        M_opt = optical_MTF_Barten1999(u, sigma)

        u = np.tile(u, (6, 1))
        sigma = np.tile(sigma, (6, 1))
        M_opt = np.tile(M_opt, (6, 1))
        np.testing.assert_almost_equal(
            optical_MTF_Barten1999(u, sigma), M_opt, decimal=7
        )

        u = np.reshape(u, (2, 3, 3))
        sigma = np.reshape(sigma, (2, 3, 3))
        M_opt = np.reshape(M_opt, (2, 3, 3))
        np.testing.assert_almost_equal(
            optical_MTF_Barten1999(u, sigma), M_opt, decimal=7
        )

    @ignore_numpy_errors
    def test_nan_optical_MTF_Barten1999(self):
        """
        Test :func:`colour.contrast.barten1999.optical_MTF_Barten1999`
        definition nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            optical_MTF_Barten1999(np.array(case), np.array(case))


class TestPupilDiameterBarten1999(unittest.TestCase):
    """
    Define :func:`colour.contrast.barten1999.pupil_diameter_Barten1999`
    definition unit tests methods.
    """

    def test_pupil_diameter_Barten1999(self):
        """
        Test :func:`colour.contrast.barten1999.pupil_diameter_Barten1999`
        definition.
        """

        np.testing.assert_almost_equal(
            pupil_diameter_Barten1999(20, 60), 2.272517118855717, decimal=7
        )

        np.testing.assert_almost_equal(
            pupil_diameter_Barten1999(0.2, 600), 2.272517118855717, decimal=7
        )

        np.testing.assert_almost_equal(
            pupil_diameter_Barten1999(20, 60, 30), 2.459028745178825, decimal=7
        )

    def test_n_dimensional_pupil_diameter_Barten1999(self):
        """
        Test :func:`colour.contrast.barten1999.pupil_diameter_Barten1999`
        definition n-dimensional support.
        """

        L = np.array([0.2, 20, 100])
        X_0 = np.array([60, 120, 240])
        Y_0 = np.array([60, 30, 15])
        d = pupil_diameter_Barten1999(L, X_0, Y_0)

        L = np.tile(L, (6, 1))
        X_0 = np.tile(X_0, (6, 1))
        d = np.tile(d, (6, 1))
        np.testing.assert_almost_equal(
            pupil_diameter_Barten1999(L, X_0, Y_0), d, decimal=7
        )

        L = np.reshape(L, (2, 3, 3))
        X_0 = np.reshape(X_0, (2, 3, 3))
        d = np.reshape(d, (2, 3, 3))
        np.testing.assert_almost_equal(
            pupil_diameter_Barten1999(L, X_0, Y_0), d, decimal=7
        )

    @ignore_numpy_errors
    def test_nan_pupil_diameter_Barten1999(self):
        """
        Test :func:`colour.contrast.barten1999.pupil_diameter_Barten1999`
        definition nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            pupil_diameter_Barten1999(
                np.array(case), np.array(case), np.array(case)
            )


class TestSigmaBarten1999(unittest.TestCase):
    """
    Define :func:`colour.contrast.barten1999.sigma_Barten1999` definition unit
    tests methods.
    """

    def test_sigma_Barten1999(self):
        """Test :func:`colour.contrast.barten1999.sigma_Barten1999` definition."""

        np.testing.assert_almost_equal(
            sigma_Barten1999(0.5 / 60, 0.08 / 60, 2.1),
            0.008791157173231,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            sigma_Barten1999(0.75 / 60, 0.08 / 60, 2.1),
            0.012809761902549,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            sigma_Barten1999(0.5 / 60, 0.16 / 60, 2.1),
            0.010040141654601,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            sigma_Barten1999(0.5 / 60, 0.08 / 60, 2.5),
            0.008975274678558,
            decimal=7,
        )

    def test_n_dimensional_sigma_Barten1999(self):
        """
        Test :func:`colour.contrast.barten1999.sigma_Barten1999` definition
        n-dimensional support.
        """

        sigma_0 = np.array([0.25 / 60, 0.5 / 60, 0.75 / 60])
        C_ab = np.array([0.04 / 60, 0.08 / 60, 0.16 / 60])
        d = np.array([2.1, 2.5, 5.0])
        sigma = sigma_Barten1999(sigma_0, C_ab, d)

        sigma_0 = np.tile(sigma_0, (6, 1))
        C_ab = np.tile(C_ab, (6, 1))
        sigma = np.tile(sigma, (6, 1))
        np.testing.assert_almost_equal(
            sigma_Barten1999(sigma_0, C_ab, d), sigma, decimal=7
        )

        sigma_0 = np.reshape(sigma_0, (2, 3, 3))
        C_ab = np.reshape(C_ab, (2, 3, 3))
        sigma = np.reshape(sigma, (2, 3, 3))
        np.testing.assert_almost_equal(
            sigma_Barten1999(sigma_0, C_ab, d), sigma, decimal=7
        )

    @ignore_numpy_errors
    def test_nan_sigma_Barten1999(self):
        """
        Test :func:`colour.contrast.barten1999.sigma_Barten1999`
        definition nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            sigma_Barten1999(np.array(case), np.array(case), np.array(case))


class TestRetinalIlluminanceBarten1999(unittest.TestCase):
    """
    Define :func:`colour.contrast.barten1999.retinal_illuminance_Barten1999`
    definition unit tests methods.
    """

    def test_retinal_illuminance_Barten1999(self):
        """
        Test :func:`colour.contrast.barten1999.retinal_illuminance_Barten1999`
        definition.
        """

        np.testing.assert_almost_equal(
            retinal_illuminance_Barten1999(20, 2.1, True),
            66.082316060529919,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            retinal_illuminance_Barten1999(20, 2.5, True),
            91.815644777503664,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            retinal_illuminance_Barten1999(20, 2.1, False),
            69.272118011654939,
            decimal=7,
        )

    def test_n_dimensional_retinal_illuminance_Barten1999(self):
        """
        Test :func:`colour.contrast.barten1999.retinal_illuminance_Barten1999`
        definition n-dimensional support.
        """

        L = np.array([0.2, 20, 100])
        d = np.array([2.1, 2.5, 5.0])
        E = retinal_illuminance_Barten1999(L, d)

        L = np.tile(L, (6, 1))
        d = np.tile(d, (6, 1))
        E = np.tile(E, (6, 1))
        np.testing.assert_almost_equal(
            retinal_illuminance_Barten1999(L, d), E, decimal=7
        )

        L = np.reshape(L, (2, 3, 3))
        d = np.reshape(d, (2, 3, 3))
        E = np.reshape(E, (2, 3, 3))
        np.testing.assert_almost_equal(
            retinal_illuminance_Barten1999(L, d), E, decimal=7
        )

    @ignore_numpy_errors
    def test_nan_retinal_illuminance_Barten1999(self):
        """
        Test :func:`colour.contrast.barten1999.retinal_illuminance_Barten1999`
        definition nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            retinal_illuminance_Barten1999(np.array(case), np.array(case))


class TestMaximumAngularSizeBarten1999(unittest.TestCase):
    """
    Define :func:`colour.contrast.barten1999.maximum_angular_size_Barten1999`
    definition unit tests methods.
    """

    def test_maximum_angular_size_Barten1999(self):
        """
        Test :func:`colour.contrast.barten1999.\
maximum_angular_size_Barten1999` definition.
        """

        np.testing.assert_almost_equal(
            maximum_angular_size_Barten1999(4, 60, 12, 15),
            3.572948005052482,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            maximum_angular_size_Barten1999(8, 60, 12, 15),
            1.851640199545103,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            maximum_angular_size_Barten1999(4, 120, 12, 15),
            3.577708763999663,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            maximum_angular_size_Barten1999(4, 60, 24, 15),
            3.698001308168194,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            maximum_angular_size_Barten1999(4, 60, 12, 30),
            6.324555320336758,
            decimal=7,
        )

    def test_n_dimensional_maximum_angular_size_Barten1999(self):
        """
        Test :func:`colour.contrast.barten1999.\
maximum_angular_size_Barten1999` definition n-dimensional support.
        """

        u = np.array([4, 8, 12])
        X_0 = np.array([60, 120, 240])
        X_max = np.array([12, 14, 16])
        N_max = np.array([15, 20, 25])
        X = maximum_angular_size_Barten1999(u, X_0, X_max, N_max)

        u = np.tile(u, (6, 1))
        X_0 = np.tile(X_0, (6, 1))
        X = np.tile(X, (6, 1))
        np.testing.assert_almost_equal(
            maximum_angular_size_Barten1999(u, X_0, X_max, N_max), X, decimal=7
        )

        u = np.reshape(u, (2, 3, 3))
        X_0 = np.reshape(X_0, (2, 3, 3))
        X = np.reshape(X, (2, 3, 3))
        np.testing.assert_almost_equal(
            maximum_angular_size_Barten1999(u, X_0, X_max, N_max), X, decimal=7
        )

    @ignore_numpy_errors
    def test_nan_maximum_angular_size_Barten1999(self):
        """
        Test :func:`colour.contrast.barten1999.\
maximum_angular_size_Barten1999` definition nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            maximum_angular_size_Barten1999(
                np.array(case), np.array(case), np.array(case), np.array(case)
            )


class TestContrastSensitivityFunctionBarten1999(unittest.TestCase):
    """
    Define :func:`colour.contrast.barten1999.\
contrast_sensitivity_function_Barten1999` definition unit tests methods.
    """

    def test_contrast_sensitivity_function_Barten1999(self):
        """
        Test :func:`colour.contrast.barten1999.\
contrast_sensitivity_function_Barten1999` definition.
        """

        np.testing.assert_almost_equal(
            contrast_sensitivity_function_Barten1999(
                u=4,
                sigma=0.01,
                E=65,
                X_0=60,
                X_max=12,
                Y_0=60,
                Y_max=12,
                p=1.2 * 10**6,
            ),
            352.761342126727020,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            contrast_sensitivity_function_Barten1999(
                u=8,
                sigma=0.01,
                E=65,
                X_0=60,
                X_max=12,
                Y_0=60,
                Y_max=12,
                p=1.2 * 10**6,
            ),
            177.706338840717340,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            contrast_sensitivity_function_Barten1999(
                u=4,
                sigma=0.02,
                E=65,
                X_0=60,
                X_max=12,
                Y_0=60,
                Y_max=12,
                p=1.2 * 10**6,
            ),
            320.872401634215750,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            contrast_sensitivity_function_Barten1999(
                u=4,
                sigma=0.01,
                E=130,
                X_0=60,
                X_max=12,
                Y_0=60,
                Y_max=12,
                p=1.2 * 10**6,
            ),
            455.171315756946400,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            contrast_sensitivity_function_Barten1999(
                u=4,
                sigma=0.01,
                E=65,
                X_0=120,
                X_max=12,
                Y_0=60,
                Y_max=12,
                p=1.2 * 10**6,
            ),
            352.996281545740660,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            contrast_sensitivity_function_Barten1999(
                u=4,
                sigma=0.01,
                E=65,
                X_0=60,
                X_max=24,
                Y_0=60,
                Y_max=12,
                p=1.2 * 10**6,
            ),
            358.881580104493650,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            contrast_sensitivity_function_Barten1999(
                u=4,
                sigma=0.01,
                E=65,
                X_0=240,
                X_max=12,
                Y_0=60,
                Y_max=12,
                p=1.2 * 10**6,
            ),
            contrast_sensitivity_function_Barten1999(
                u=4,
                sigma=0.01,
                E=65,
                X_0=60,
                X_max=12,
                Y_0=240,
                Y_max=12,
                p=1.2 * 10**6,
            ),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            contrast_sensitivity_function_Barten1999(
                u=4,
                sigma=0.01,
                E=65,
                X_0=60,
                X_max=12,
                Y_0=60,
                Y_max=12,
                p=1.4 * 10**6,
            ),
            374.791328640476140,
            decimal=7,
        )

    def test_n_dimensional_contrast_sensitivity_function_Barten1999(self):
        """
        Test :func:`colour.contrast.barten1999.\
contrast_sensitivity_function_Barten1999` definition n-dimensional support.
        """

        u = np.array([4, 8, 12])
        sigma = np.array([0.01, 0.02, 0.04])
        E = np.array([0.65, 90, 1500])
        X_0 = np.array([60, 120, 240])
        S = contrast_sensitivity_function_Barten1999(
            u=u, sigma=sigma, E=E, X_0=X_0
        )

        u = np.tile(u, (6, 1))
        E = np.tile(E, (6, 1))
        S = np.tile(S, (6, 1))
        np.testing.assert_almost_equal(
            contrast_sensitivity_function_Barten1999(
                u=u, sigma=sigma, E=E, X_0=X_0
            ),
            S,
            decimal=7,
        )

        u = np.reshape(u, (2, 3, 3))
        E = np.reshape(E, (2, 3, 3))
        S = np.reshape(S, (2, 3, 3))
        np.testing.assert_almost_equal(
            contrast_sensitivity_function_Barten1999(
                u=u, sigma=sigma, E=E, X_0=X_0
            ),
            S,
            decimal=7,
        )

    @ignore_numpy_errors
    def test_nan_contrast_sensitivity_function_Barten1999(self):
        """
        Test :func:`colour.contrast.barten1999.\
contrast_sensitivity_function_Barten1999` definition nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            contrast_sensitivity_function_Barten1999(
                u=np.array(case),
                sigma=np.array(case),
                E=np.array(case),
                X_0=np.array(case),
            )


if __name__ == "__main__":
    unittest.main()



"""
NTSC Colourspaces
=================

Defines the *NTSC* colourspaces:

-   :attr:`colour.models.RGB_COLOURSPACE_NTSC1953`.
-   :attr:`colour.models.RGB_COLOURSPACE_NTSC1987`.

References
----------
-   :cite:`InternationalTelecommunicationUnion1998a` : International
    Telecommunication Union. (1998). Recommendation ITU-R BT.470-6 -
    CONVENTIONAL TELEVISION SYSTEMS (pp. 1-36).
    http://www.itu.int/dms_pubrec/itu-r/rec/bt/\
R-REC-BT.470-6-199811-S!!PDF-E.pdf
-   :cite:`SocietyofMotionPictureandTelevisionEngineers2004a` : Society of
    Motion Picture and Television Engineers. (2004). RP 145:2004: SMPTE C Color
    Monitor Colorimetry. In RP 145:2004: Vol. RP 145:200. The Society of Motion
    Picture and Television Engineers. doi:10.5594/S9781614821649
"""

from __future__ import annotations

from colour.hints import NDArray
from colour.models.rgb import RGB_Colourspace
from colour.models.rgb.datasets.itur_bt_470 import (
    PRIMARIES_BT470_525,
    CCS_WHITEPOINT_BT470_525,
    WHITEPOINT_NAME_BT470_525,
    MATRIX_BT470_525_TO_XYZ,
    MATRIX_XYZ_TO_BT470_525,
    RGB_COLOURSPACE_BT470_525,
)
from colour.models.rgb.datasets.smpte_c import (
    PRIMARIES_SMPTE_C,
    WHITEPOINT_NAME_SMPTE_C,
    CCS_WHITEPOINT_SMPTE_C,
    MATRIX_SMPTE_C_TO_XYZ,
    MATRIX_XYZ_TO_SMPTE_C,
    RGB_COLOURSPACE_SMPTE_C,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "PRIMARIES_NTSC1953",
    "WHITEPOINT_NAME_NTSC1953",
    "CCS_WHITEPOINT_NTSC1953",
    "MATRIX_NTSC1953_TO_XYZ",
    "MATRIX_XYZ_TO_NTSC1953",
    "RGB_COLOURSPACE_NTSC1953",
    "PRIMARIES_NTSC1987",
    "WHITEPOINT_NAME_NTSC1987",
    "CCS_WHITEPOINT_NTSC1987",
    "MATRIX_NTSC1987_TO_XYZ",
    "MATRIX_XYZ_TO_NTSC1987",
    "RGB_COLOURSPACE_NTSC1987",
]

PRIMARIES_NTSC1953: NDArray = PRIMARIES_BT470_525
"""*NTSC (1953)* colourspace primaries."""

WHITEPOINT_NAME_NTSC1953: str = WHITEPOINT_NAME_BT470_525
"""*NTSC (1953)* colourspace whitepoint name."""

CCS_WHITEPOINT_NTSC1953: NDArray = CCS_WHITEPOINT_BT470_525
"""*NTSC (1953)* colourspace whitepoint chromaticity coordinates."""

MATRIX_NTSC1953_TO_XYZ: NDArray = MATRIX_BT470_525_TO_XYZ
"""*NTSC (1953)* colourspace to *CIE XYZ* tristimulus values matrix."""

MATRIX_XYZ_TO_NTSC1953: NDArray = MATRIX_XYZ_TO_BT470_525
"""*CIE XYZ* tristimulus values to *NTSC (1953)* colourspace matrix."""

RGB_COLOURSPACE_NTSC1953: RGB_Colourspace = RGB_Colourspace(
    "NTSC (1953)",
    PRIMARIES_NTSC1953,
    CCS_WHITEPOINT_NTSC1953,
    WHITEPOINT_NAME_NTSC1953,
    MATRIX_NTSC1953_TO_XYZ,
    MATRIX_XYZ_TO_NTSC1953,
    RGB_COLOURSPACE_BT470_525.cctf_encoding,
    RGB_COLOURSPACE_BT470_525.cctf_decoding,
)
RGB_COLOURSPACE_NTSC1953.__doc__ = """
*NTSC (1953)* colourspace.

References
----------
:cite:`InternationalTelecommunicationUnion1998a`
"""

PRIMARIES_NTSC1987: NDArray = PRIMARIES_SMPTE_C
"""*NTSC (1987)* colourspace primaries."""

WHITEPOINT_NAME_NTSC1987: str = WHITEPOINT_NAME_SMPTE_C
"""*NTSC (1987)* colourspace whitepoint name."""

CCS_WHITEPOINT_NTSC1987: NDArray = CCS_WHITEPOINT_SMPTE_C
"""*NTSC (1987)* colourspace whitepoint chromaticity coordinates."""

MATRIX_NTSC1987_TO_XYZ: NDArray = MATRIX_SMPTE_C_TO_XYZ
"""*NTSC (1987)* colourspace to *CIE XYZ* tristimulus values matrix."""

MATRIX_XYZ_TO_NTSC1987: NDArray = MATRIX_XYZ_TO_SMPTE_C
"""*CIE XYZ* tristimulus values to *NTSC (1987)* colourspace matrix."""

RGB_COLOURSPACE_NTSC1987: RGB_Colourspace = RGB_Colourspace(
    "NTSC (1987)",
    PRIMARIES_NTSC1987,
    CCS_WHITEPOINT_NTSC1987,
    WHITEPOINT_NAME_NTSC1987,
    MATRIX_NTSC1987_TO_XYZ,
    MATRIX_XYZ_TO_NTSC1987,
    RGB_COLOURSPACE_SMPTE_C.cctf_encoding,
    RGB_COLOURSPACE_SMPTE_C.cctf_decoding,
)
RGB_COLOURSPACE_NTSC1987.__doc__ = """
*NTSC (1987)* colourspace.

References
----------
:cite:`SocietyofMotionPictureandTelevisionEngineers2004a`
"""



"""
Standardized Residual Sum of Squares (STRESS) Index
===================================================

Defines the *Kruskal's Standardized Residual Sum of Squares (:math:`STRESS`)*
index:

-   :func:`colour.index_stress_Garcia2007`: :math:`STRESS` index computation
    according to *García, Huertas, Melgosa and Cui (2007)* method.
-   :func:`colour.index_stress`: *Lightness* :math:`STRESS` index computation
    according to given method.

References
----------
-   :cite:`Garcia2007` : García, P. A., Huertas, R., Melgosa, M., & Cui, G.
    (2007). Measurement of the relationship between perceived and computed
    color differences. Journal of the Optical Society of America A, 24(7),
    1823. doi:10.1364/JOSAA.24.001823
"""

from __future__ import annotations

import numpy as np

from colour.hints import FloatingOrArrayLike, FloatingOrNDArray, Literal, Union
from colour.utilities import (
    CaseInsensitiveMapping,
    as_float,
    as_float_array,
    validate_method,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "index_stress_Garcia2007",
    "INDEX_STRESS_METHODS",
    "index_stress",
]


def index_stress_Garcia2007(
    d_E: FloatingOrArrayLike, d_V: FloatingOrArrayLike
) -> FloatingOrNDArray:
    """
    Compute the
    *Kruskal's Standardized Residual Sum of Squares (:math:`STRESS`)*
    index according to *García, Huertas, Melgosa and Cui (2007)* method.

    Parameters
    ----------
    d_E
        Computed colour difference array :math:`\\Delta E`.
    d_V
        Computed colour difference array :math:`\\Delta V`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        :math:`STRESS` index.

    References
    ----------
    :cite:`Garcia2007`

    Examples
    --------
    >>> d_E = np.array([2.0425, 2.8615, 3.4412])
    >>> d_V = np.array([1.2644, 1.2630, 1.8731])
    >>> index_stress_Garcia2007(d_E, d_V)  # doctest: +ELLIPSIS
    0.1211709...
    """

    d_E = as_float_array(d_E)
    d_V = as_float_array(d_V)

    F_1 = np.sum(d_E**2) / np.sum(d_E * d_V)

    stress = np.sqrt(
        np.sum((d_E - F_1 * d_V) ** 2) / np.sum(F_1**2 * d_V**2)
    )

    return as_float(stress)


INDEX_STRESS_METHODS: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "Garcia 2007": index_stress_Garcia2007,
    }
)
INDEX_STRESS_METHODS.__doc__ = """
Supported :math:`STRESS` index computation methods.

References
----------
:cite:`Garcia2007`
"""


def index_stress(
    d_E: FloatingOrArrayLike,
    d_V: FloatingOrArrayLike,
    method: Union[Literal["Garcia 2007"], str] = "Garcia 2007",
) -> FloatingOrNDArray:
    """
    Compute the
    *Kruskal's Standardized Residual Sum of Squares (:math:`STRESS`)*
    index according to given method.

    Parameters
    ----------
    d_E
        Computed colour difference array :math:`\\Delta E`.
    d_V
        Computed colour difference array :math:`\\Delta V`.
    method
        Computation method.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        :math:`STRESS` index.

    References
    ----------
    :cite:`Garcia2007`

    Examples
    --------
    >>> d_E = np.array([2.0425, 2.8615, 3.4412])
    >>> d_V = np.array([1.2644, 1.2630, 1.8731])
    >>> index_stress(d_E, d_V)  # doctest: +ELLIPSIS
    0.1211709...
    """

    method = validate_method(method, INDEX_STRESS_METHODS)

    function = INDEX_STRESS_METHODS[method]

    return function(d_E, d_V)



"""
CIE XYZ Tristimulus Values of the Hunter L,a,b Illuminants
==========================================================

Defines the *CIE XYZ* tristimulus values of the *Hunter L,a,b* illuminants
dataset for the *CIE 1931 2 Degree Standard Observer* and
*CIE 1964 10 Degree Standard Observer*.

The currently implemented data has been extracted from :cite:`HunterLab2008b`,
however you may want to use different data according to the tables given in
:cite:`HunterLab2008c`.

References
----------
-   :cite:`HunterLab2008b` : HunterLab. (2008). Hunter L,a,b Color Scale.
    http://www.hunterlab.se/wp-content/uploads/2012/11/Hunter-L-a-b.pdf
-   :cite:`HunterLab2008c` : HunterLab. (2008). Illuminant Factors in
    Universal Software and EasyMatch Coatings.
    https://support.hunterlab.com/hc/en-us/article_attachments/201437785/\
an02_02.pdf
"""

from __future__ import annotations

import numpy as np
from collections import namedtuple

from colour.hints import Tuple
from colour.utilities import CaseInsensitiveMapping

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "Illuminant_Specification_HunterLab",
    "DATA_ILLUMINANTS_HUNTERLAB_STANDARD_OBSERVER_2_DEGREE_CIE1931",
    "TVS_ILLUMINANTS_HUNTERLAB_STANDARD_OBSERVER_2_DEGREE_CIE1931",
    "DATA_ILLUMINANTS_HUNTERLAB_STANDARD_OBSERVER_10_DEGREE_CIE1964",
    "TVS_ILLUMINANTS_HUNTERLAB_STANDARD_OBSERVER_10_DEGREE_CIE1964",
    "TVS_ILLUMINANTS_HUNTERLAB",
]

Illuminant_Specification_HunterLab = namedtuple(
    "Illuminant_Specification_HunterLab", ("name", "XYZ_n", "K_ab")
)

DATA_ILLUMINANTS_HUNTERLAB_STANDARD_OBSERVER_2_DEGREE_CIE1931: Tuple = (
    ("A", np.array([109.83, 100.00, 35.55]), np.array([185.20, 38.40])),
    ("C", np.array([98.04, 100.00, 118.11]), np.array([175.00, 70.00])),
    ("D50", np.array([96.38, 100.00, 82.45]), np.array([173.51, 58.48])),
    ("D60", np.array([95.23, 100.00, 100.86]), np.array([172.47, 64.72])),
    ("D65", np.array([95.02, 100.00, 108.82]), np.array([172.30, 67.20])),
    ("D75", np.array([94.96, 100.00, 122.53]), np.array([172.22, 71.30])),
    ("FL2", np.array([98.09, 100.00, 67.53]), np.array([175.00, 52.90])),
    ("TL 4", np.array([101.40, 100.00, 65.90]), np.array([178.00, 52.30])),
    ("UL 3000", np.array([107.99, 100.00, 33.91]), np.array([183.70, 37.50])),
)

TVS_ILLUMINANTS_HUNTERLAB_STANDARD_OBSERVER_2_DEGREE_CIE1931: (
    CaseInsensitiveMapping
) = CaseInsensitiveMapping(
    {
        x[0]: Illuminant_Specification_HunterLab(*x)
        for x in DATA_ILLUMINANTS_HUNTERLAB_STANDARD_OBSERVER_2_DEGREE_CIE1931
    }
)
"""
*CIE XYZ* tristimulus values of the *Hunter L,a,b* illuminants for the
*CIE 1931 2 Degree Standard Observer*.

References
----------
:cite:`HunterLab2008b`, :cite:`HunterLab2008c`
"""

DATA_ILLUMINANTS_HUNTERLAB_STANDARD_OBSERVER_10_DEGREE_CIE1964: Tuple = (
    ("A", np.array([111.16, 100.00, 35.19]), np.array([186.30, 38.20])),
    ("C", np.array([97.30, 100.00, 116.14]), np.array([174.30, 69.40])),
    ("D50", np.array([96.72, 100.00, 81.45]), np.array([173.82, 58.13])),
    ("D60", np.array([95.21, 100.00, 99.60]), np.array([172.45, 64.28])),
    ("D65", np.array([94.83, 100.00, 107.38]), np.array([172.10, 66.70])),
    ("D75", np.array([94.45, 100.00, 120.70]), np.array([171.76, 70.76])),
    ("FL2", np.array([102.13, 100.00, 69.37]), np.array([178.60, 53.60])),
    ("TL 4", np.array([103.82, 100.00, 66.90]), np.array([180.10, 52.70])),
    ("UL 3000", np.array([111.12, 100.00, 35.21]), np.array([186.30, 38.20])),
)

TVS_ILLUMINANTS_HUNTERLAB_STANDARD_OBSERVER_10_DEGREE_CIE1964: (
    CaseInsensitiveMapping
) = CaseInsensitiveMapping(
    {
        x[0]: Illuminant_Specification_HunterLab(*x)
        for x in DATA_ILLUMINANTS_HUNTERLAB_STANDARD_OBSERVER_10_DEGREE_CIE1964
    }
)
"""
*CIE XYZ* tristimulus values of the *Hunter L,a,b* illuminants for the
*CIE 1964 10 Degree Standard Observer*.

References
----------
:cite:`HunterLab2008b`, :cite:`HunterLab2008c`
"""

TVS_ILLUMINANTS_HUNTERLAB: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "CIE 1931 2 Degree Standard Observer": (
            TVS_ILLUMINANTS_HUNTERLAB_STANDARD_OBSERVER_2_DEGREE_CIE1931
        ),
        "CIE 1964 10 Degree Standard Observer": (
            TVS_ILLUMINANTS_HUNTERLAB_STANDARD_OBSERVER_10_DEGREE_CIE1964
        ),
    }
)
TVS_ILLUMINANTS_HUNTERLAB.__doc__ = """
*CIE XYZ* tristimulus values of the *HunterLab* illuminants.

References
----------
:cite:`HunterLab2008b`, :cite:`HunterLab2008c`

Aliases:

-   'cie_2_1931': 'CIE 1931 2 Degree Standard Observer'
-   'cie_10_1964': 'CIE 1964 10 Degree Standard Observer'
"""
TVS_ILLUMINANTS_HUNTERLAB["cie_2_1931"] = TVS_ILLUMINANTS_HUNTERLAB[
    "CIE 1931 2 Degree Standard Observer"
]
TVS_ILLUMINANTS_HUNTERLAB["cie_10_1964"] = TVS_ILLUMINANTS_HUNTERLAB[
    "CIE 1964 10 Degree Standard Observer"
]
